					//*********************************************************************************
					//
					//     This program is free software; you can redistribute it and/or modify
					//     it under the terms of the GNU General Public License as published by
					//     the Free Software Foundation; either version 2 of the License, or
					//     (at your option) any later version.
					//
					//     This program is distributed in the hope that it will be useful,
					//     but WITHOUT ANY WARRANTY; without even the implied warranty of
					//     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
					//     GNU General Public License for more details.
					//
					//    You should have received a copy of the GNU General Public License along
					//    with this program; if not, write to the Free Software Foundation, Inc.,
					//    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
					//    
					//    A copy of the GNU General Public License is also available at
					//    http://www.fsf.org/licensing/licenses
					//
					//*********************************************************************************


					//*************************************************************************************
					//
					//	This Trick is based on the Sylvester Theorem for determinants, since the 
					//	determinant is used to obtain the eigenvalues:
					//
					//	Original Problem states: Find Eigenvalues of the Matrix:
					//	L_nm=sqrt(A_n) (K^t d^2L/dDAdDA K)_nm sqrt(A_m)
					//
					//	Since L is symmetric and real, we can diagonalize it using orthonormal 
					//	matrices  L = Q D Q^t, where D is the diagonal matrix that houses all
					//	eigenvalues on its diagonal.
					//
					//	If we instead look at the square of lambda we will get
					//	L*L = Q D Q^t Q D Q^t = Q D^2 Q^t, where now on the diagonal in D we have
					//	the squares of the eigenvalues. I.e. if we diagonlaize L*L and take the
					//	square root of the diagonal elements, we still obtain the eigenvalues.
					//
					//	Now comes the trick using det(I_n+A*B) = det(I_m+B*A) for rectangular 
					//	matrices A and B. Especially we have the Eigenvalue formula:
					//	det( A_(nxn)-l*I_n)==0 ->  (-l)^n det(I+A'_(nxn))==0 where
					//	A' is just -1/l*A.
					//
					//	Using the SVD we can write our original Lambda as:
					//	L   = sqrt(A_n) ( U S V^t d^2L/dDAdDA V S U^t )_nm sqrt(A_m)
					//	from which follows
					//	L*L = sqrt(A_n) ( U M T M U^t )_nm sqrt(A_m)
					//	with the definitions
					//	M= S V^t d^2L/dDAdDA V S and T= U^t diag[A] U and U'_nm=sqrt(A_n)U_nm
					//
					//	Combining the prevoius to paragraphs we get:
					//	det(A-l*I) = (-l)^n det(I_Nw + (-1/l) U' M T M U'^t )
					//		   = (-l)^n det(I_Nt + (-1/l) M T M U'^t U' )
					//		   = (-l)^n det(I_Nt + (-1/l) M T M T )
					//		   = det ( MTMT - l I_Nt ) == 0
					//
					//	At the same time we can also use 
					//	det(A-l*I) = (-l)^n det(I_Nw + (-1/l) U' M T M U'^t )
					//		   = (-l)^n det(I_Nt + (-1/l) U'^t U' M T M  )
					//		   = (-l)^n det(I_Nt + (-1/l) T M T M )
					//		   = det ( TMTM - l I_Nt ) == 0
					//
					//	So that we find that the Eigenvalues of T M T M = M T M T.
					//
					//	In earlier implementations the formula L*L= M T T M appears,
					//	which however does not coincide with M T M T since M T != T M
					//	(The product of two symmetric matrices is not necessarily symmetric!)
					//
					//	Also one might think that because MT and TM have the same Eigenvalues
					//	we could also write MTTM instead of MTMT. But although the Eigenvalues
					//	are the same the Eigenvectors do not necessarily coincide:
					//	MT -> Q D_MT Q^t TM-> U D_TM U^t ==> MTTM -> Q D_MT Q^t U D_TM U^t
					//	since Q^t U != 1 this does not help us further!
					//
					//	NOW THAT WE ARE USING THE ALGLIB ROUTINES FOR THE EIGENVALUES OF
					//	NON SYMMETRIC MATRICES WITH MPFR WE CAN ATTACK LAMBDA ITSELF:  
					//
					//	L   = sqrt(A_n) ( U S V^t d^2L/dDAdDA V S U^t )_nm sqrt(A_m)
					//
					//	det(L-l*I) = (-l)^n det(I_Nw + (-1/l) U' M  U'^t )
					//		   = (-l)^n det(I_Nt + (-1/l) M U'^t U' )
					//		   = (-l)^n det(I_Nt + (-1/l) M T )
					//		   = det ( MT - l I_Nt ) == 0
					//
					//	WE NEED TO GET THE EIGENVALUES OF MT ONLY
					//
					//*************************************************************************************

					gslgmp_vector* LEvalr;

					Lambda=gslgmp_matrix_calloc(NSING,NSING);
					
					LEvalr=gslgmp_vector_calloc(NSING);
					
					T2=gslgmp_matrix_calloc(NSING,NSING);
					M2=gslgmp_matrix_calloc(NSING,NSING);
					
					//*********************************************************
					//	Construct M
					//*********************************************************

					gslgmp_matrix* XisMM;
					XisMM=gslgmp_matrix_calloc(NSING,NSING);
				
					gslgmp_matrix* VsM;
					VsM=gslgmp_matrix_calloc(TWIDTH,NSING);
				
					gslgmp_matrix* ddLddDAM;
					ddLddDAM=gslgmp_matrix_calloc(TWIDTH,TWIDTH);
		
					gslgmp_matrix_memcpy(XisMM,XisM);
					gslgmp_matrix_memcpy(VsM,Vs);
		
					for(int dt1=0;dt1<TWIDTH;dt1++){
					for(int dt2=0;dt2<TWIDTH;dt2++){
						gslgmp_matrix_set_ui(ddLddDAM,dt1,dt2,0);
						if(dt1==dt2){
							gslgmp_vector_get(CEval,dt1,ptmp1);
							mpfr_ui_div(ptmp2,1,ptmp1,GMP_RNDN);
							gslgmp_matrix_set(ddLddDAM,dt1,dt2,ptmp2);
						}
					}}	
					
					gmp_mult5m(XisMM,0,VsM,1,ddLddDAM,0,VsM,0,XisMM,0,M2);
				
					gslgmp_matrix_free(XisMM);
					gslgmp_matrix_free(VsM);
					gslgmp_matrix_free(ddLddDAM);
				
					//*********************************************************
					//	Construct T
					//*********************************************************

					gslgmp_matrix* UMATM;
					UMATM=gslgmp_matrix_calloc(Us->size1,Us->size2);
				
					gslgmp_matrix_memcpy(UMATM,Us);
				
					gmp_multmDIAGm(UMATM,1,Adw,UMATM,0,T2);
				
					gslgmp_matrix_free(UMATM);
		
					//*********************************************************
					//	Calculate MT
					//*********************************************************

					gmp_mult2m(M2,0,T2,0,Lambda);
					
					//*********************************************************
					//	Calculate Eigenvalues of MT using ALGLIB!
					//*********************************************************

					ap::template_1d_array< amp::ampf > ALG_Evalr;
					ap::template_1d_array< amp::ampf > ALG_Evali;
					
					ap::template_2d_array< amp::ampf > ALG_Evecr;
					ap::template_2d_array< amp::ampf > ALG_Eveci;

					ap::template_2d_array< amp::ampf > ALG_Lambda;

					ALG_Evalr.setlength(NSING);
					ALG_Evali.setlength(NSING);
					ALG_Lambda.setlength(NSING,NSING);
					ALG_Evecr.setlength(NSING,NSING);
					ALG_Eveci.setlength(NSING,NSING);

					for (int i=0;i<NSING;i++)
					for (int j=0;j<NSING;j++){
						
						gslgmp_matrix_get(Lambda,i,j,ptmp1);
						mpfr_set(ALG_Lambda(i,j).getWritePtr(),ptmp1,GMP_RNDN);
						
					}

					int Alen=NSING;
					int Avecs=0;

					evd::rmatrixevd(ALG_Lambda, Alen, Avecs, ALG_Evalr, ALG_Evali, ALG_Evecr, ALG_Eveci);
				
					//*********************************************************
					//	Since original matrix had no complex eigenvalues
					//	we check for consistency that also now Im[EV]=0
					//*********************************************************

					LEvalr=gslgmp_vector_calloc(Lambda->size1);
					
					mpfr_t myprec;
					mpfr_init2(myprec,PRECBITS);
					mpfr_ui_pow_ui(myprec,2,PRECBITS,GMP_RNDN);
					mpfr_ui_div(myprec,1,myprec,GMP_RNDN);
 					mpfr_mul_ui(myprec,myprec,4,GMP_RNDN);
					
					for(int i=0;i<NSING;i++){
						
						if( mpfr_greater_p( ALG_Evali(i).getReadPtr(), myprec ) ){
							fprintf(stderr, "Caution eigenvalue:%d has imaginary part larger than %e, Possible Precision problem\n",i, mpfr_get_d(myprec,GMP_RNDN) );
						}
						
						mpfr_set( LEvalr->data[i],ALG_Evalr(i).getReadPtr(),GMP_RNDN);
						
					}
					
					fprintf(stderr, "**********************************************************\n");
					fprintf(stderr, "**** EIGENVALUES OF THE LAMBDA MATRIX             ********\n");
					fprintf(stderr, "**********************************************************\n");
										
					gmpPVec(LEvalr);
					
					//*********************************************************
					//	Since Lambda is susceptible for leaving the
					//	range of precision check if NAN or inf
					//*********************************************************
				
					for (int i=0;i<NSING;i++){
				
						mpfr_set(ptmp1,LEvalr->data[i],GMP_RNDN);
						if(mpfr_nan_p(ptmp1)!=0||mpfr_inf_p(ptmp1)!=0) {fprintf(stderr,"Inf/Nan encountered in Eigenvalue calculation:%e\n",mpfr_get_d(ptmp1,GMP_RNDN)); mpfr_set_d(ptmp1,0.,GMP_RNDN);};
						//gslgmp_vector_set(LEval,i,ptmp1);
					}
		
					//*********************************************************
					//	Calculate the argument of the exponential in the 
					//	Probability formula (3.19) for the current alpha
					//	0.5*Sum_k Log[alpha/(alpha+lambda_k)]+alpha*S-L
					//*********************************************************

					//*********************************************************
					//	FROM NICKEL: PROBABILITY TERM HAS THE FORM (SEE UNITS):
					//
					//	SUM_i Log[ ALPHA*DW/(ALPHA*DW+LAMBDA_i) ]
					//
					//	SINCE WE ARE USING THE SYLVESTER TRICK
					//	I ONLY SCALE WITH A SINGLE DW
					//*********************************************************
		
					mpfr_t asumtrick;
					mpfr_init2(asumtrick,PRECBITS);
					mpfr_set_ui(asumtrick,0,GMP_RNDN);
					
					for(int dw=0;dw<NSING;dw++){
							
						mpfr_sub(ptmp1,omega->data[1],omega->data[0],GMP_RNDN);
					
						mpfr_set_d(ptmp2,alpha,GMP_RNDN);
						mpfr_mul(ptmp1,ptmp1,ptmp2,GMP_RNDN);

						mpfr_add(ptmp2,ptmp1,LEvalr->data[dw],GMP_RNDN);
						mpfr_div(ptmp3,ptmp1,ptmp2,GMP_RNDN);
						mpfr_log(ptmp1,ptmp3,GMP_RNDN);
				
						mpfr_div_ui(ptmp1,ptmp1,2,GMP_RNDN);
						mpfr_add(asumtrick,asumtrick,ptmp1,GMP_RNDN);
					}

						
					if(acnt<ANUM) { 
						
						mpfr_mul_d(ptmp1,Sval,alpha,GMP_RNDN);
						mpfr_sub(ptmp1,ptmp1,Lval,GMP_RNDN);
						mpfr_add(ptmp1,ptmp1,asumtrick,GMP_RNDN);
						mpfr_set(LOGPROB->data[acnt],ptmp1,GMP_RNDN);
						
						}

					fprintf(stderr, "**********************************************************\n");
					fprintf(stderr, "**** ASUM %.5e LOGPROB %.5e              ********\n",mpfr_get_d(asumtrick,GMP_RNDN),mpfr_get_d(LOGPROB->data[acnt],GMP_RNDN));
					fprintf(stderr, "**********************************************************\n");

					mpfr_clear(asumtrick);
					gslgmp_matrix_free(Lambda);
					gslgmp_vector_free(LEvalr);
					gslgmp_matrix_free(M2);
					gslgmp_matrix_free(T2);
