//*********************************************************************************
//	ExtMEM v3.11
//
//	THIS PROGRAM IS AN IMPLEMENTATION OF THE MAXIMUM ENTROPY METHOD
//	FOR THE NUMERICAL INVERSION OF THE TYPICALLY ILL DEFINED TASK
//
//	DATA[t] = Int_-inf^inf [ K[w,t] * SPEC[w] ]dw
//
//	WHERE THE GOAL IS TO EXTRACT THE SPECTRAL FUNCTION SPEC[w] FROM
//	THE PROVIDED LATTICE DATA DATA[t] USING A PREVIOUSLY SPECIFIED
//	KERNEL FUNCTION K[w,t].
//
//	AUTHOR: Dr. Alexander Karl Rothkopf	DATE: June 20th 2012
//
//	References:
//
//	Improved Maximum Entropy Analysis with an Extended Search Space
//	A.Rothkopf (Bielefeld. U)
//	e-Print: arXiv:1110.6285
//	Url: http://arxiv.org/abs/arXiv:1110.6285
//
//	I AM GRATEFUL FOR THE HELPFUL COMMENTS FROM T.HATSUDA, S.SASAKI, 
//	K. SASAKI, O. KACZMAREK AND H.T. DING
//
//*********************************************************************************

//*********************************************************************************
//
//     Copyright (C) 2012  Dr. Alexander Rothkopf
//
//     This program is free software; you can redistribute it and/or modify
//     it under the terms of the GNU General Public License as published by
//     the Free Software Foundation; either version 2 of the License, or
//     (at your option) any later version.
//
//     This program is distributed in the hope that it will be useful,
//     but WITHOUT ANY WARRANTY; without even the implied warranty of
//     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//     GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License along
//    with this program; if not, write to the Free Software Foundation, Inc.,
//    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//    
//    A copy of the GNU General Public License is also available at
//    http://www.fsf.org/licensing/licenses
//
//*********************************************************************************

//*********************************************************************************
//	This program uses adapted code from two libraries:
//
//	1. The GNU Scientific Library:
//
//		SINGULAR VALUE DECOMPOSITION
//		EIGENSYSTEMS OF SYMMETRIC MATRICES
//		MATRIX INVERSION BASED ON QR DECOMPOSITION
//		SOLVER FOR Ax=b BASED ON QR DECOMPOSITION
//
//	2. The ALGLIB Library (MPFR part 2.6.):
//
//		LBFGS MINIMIZER
//		LEVENBERG MARQUARD + LBFGS PRECONDITIONER MINIMIZER
//		EIGENVALUES FOR NON_SYMMETRIC MATRICES
//
//	POSSIBLE OPTIMIZATION: MOVE ALL GSL ROUTINES TO THEIR ALGLIB COUNTERPART
//
//*********************************************************************************


//*********************************************************************************
//	Standard includes for math, input, output and nice
//*********************************************************************************

    #include <stdio.h>
    #include <stdlib.h>
    #include <math.h>
    #include <ctime>

//*********************************************************************************
//	Arbitrary Precision Datatypes
//*********************************************************************************

    #include <mpfr.h>

    #define PI (3.141592653589793)
    #define GNUPR(XX)  (((XX*XX)<1e-60)?0.:XX)
    
//*********************************************************************************
//	The only two global variables
//********************************************************************************

    int PRECBITS;			//GLOBAL PRECISION FOR GSLGMP ROUTINES
    int Precision;			//GLOBAL PRECISION FOR ALGLIB ROUTINES
    int PRECISION_SET=0;		//CHECK WHETHER PRECISION HAS ALREADY BEEN SET
    
//*********************************************************************************
//	THE MODIFIED ALGLIB FILES 
//
//	INSTEAD OF THE TEMPLATE STRUCTURE WE ARE USING A GLOBAL VARIABLE
//	CALLED "Precision" TO SET THE PRECISION OF THE MPFR ARITHMETIC
//	IN THE EIGENVALUE SOLVERS AND THE MINIMIZER ROUTINES. 
//
//*********************************************************************************

    #include "./ALGLIB_ROUTINES/ap.h"
    #include "./ALGLIB_ROUTINES/amp.h"

    #include "./ALGLIB_ROUTINES/minlbfgs.h"
    #include "./ALGLIB_ROUTINES/minlm.h"
    #include "./ALGLIB_ROUTINES/evd.h"    
   
    #include "./ALGLIB_ROUTINES/ap.cpp"
    #include "./ALGLIB_ROUTINES/amp.cpp"
   
   
//*********************************************************************************
//	Parameters for the Program
//*********************************************************************************

    #define DECORR 1		// TRANSFORM THE DATA AND KERNEL INTO THE DECORRELATED BASIS

    #define SETCORRDIAG 0	// NEGLECT THE OFFDIAGONAL ENTRIES IN THE CORRELATION MATRIX

    #define LTRICK 1		// 0: DIRECT LAMDA EV 1: LAMDA EV VIA SYLVESTER AND NON-SYMM SOLVER   

    #define REDOMEM 0		// 0: USUAL MEM 1: LOAD VALUES FROM MEM_spec_rec_alpha AND MEM_DCDXDATA to repeat the ERROR AND PROBABILITY ESTIMATION

//*********************************************************************************
//	MISC OPTIONS FOR TESTING THE PROGRAM
//*********************************************************************************

    #define ITERPRIOR 0		// USE THE RECONSTRUCTED IMAGE AS A PRIOR FOR AN ADDITIONAL RECONSTRUCTION RUN

    #define GLOBALPRIORREP 2	// HOW OFTEN TO REPEAT THE ABOVE STRATEGY (ITERPRIOR)

    #define BFGSPREC 0		// STOPPING CRITERION FOR THE BFGS METHOD (INTERNAL)

    #define VERBOSE 1		// SHOW A LOT OF OUTPUT
    #define DVERBOSE 0		// SHOW EVEN MORE OUTPUT

    #define WITHDW 1		// MULTIPLY THE DW TERMS IN THE DERIVATIVE INSIDE THE LBFGS MINIMIZER

    #define GAUSSCHK 0		// CALCULATE AND OUTPUT SKEWNESS AND KURTOSIS DURING DATA ACQUISITION

    #define CONDNSING (0.)	// HOW SMALL MUST A SINGULAR VALUE BE TO BE TAKEN AS ZERO

    #define DEBUGTST 0		// RUN A TESTSUITE FOR THE SUBPROCESSES IN THE PROGRAM

    #define SVDOUT 1
    
    #define GUILARGE 0

//*********************************************************************************
//	Definitions for the High Precision gslgmp datatypes
//*********************************************************************************

    #include "MEM.h"
   
//*********************************************************************************
//	GUI Part
//*********************************************************************************

    #if (WITHGUI==1)
    
    #include "./GUI/GUI.h"
    
    #include "./GUI/GUI.cpp"
 
    IMPLEMENT_APP(MEMGui)

    #endif

//*********************************************************************************
//	Auxiliary routines
//*********************************************************************************

    #include "./AUXILIARY_ROUTINES/APPENDDW.c"		// Function to append correct frequency measure
    #include "./AUXILIARY_ROUTINES/DATA_ACQ.c"		// Data acquisition routines and mock data 

    #if (DEBUGTST==1)
	#include "./AUXILIARY_ROUTINES/TEST_SUBSYSTEMS.c"
    #endif

//*********************************************************************************
//	MAIN ROUTINE
//*********************************************************************************
    #if (WITHGUI==0)
    int main(int argc,char *argv[])
    {
    #else
    int MEM_main(int argc,char *argv[], ThreadInfo *TInfo)
    {
    #endif
    

	fprintf(stderr, "\n**********************************************************\n");
	fprintf(stderr, "***   THIS IS PROGRAM VERSION v3.1                     ***\n");
	fprintf(stderr, "**********************************************************\n");
	
	//*************************************************************************
	//	DEFINE THE PARAMETER VARIABLES IN THE CODE
	//*************************************************************************
	
	double BETA,ALPHAMAX,ALPHAMIN,MFAC,WMAX,WMIN,MOCKNOISE,HPSTART,HPEND,ASTART,AMIN,AMAX;
	int XMAX,NX,TREADMAX,NT,NOMEGA,ERRNUM,NUMCONF,FOFFSET,CHEATDP,NTMIN,NTMAX,METHOD;
	int TWIDTH,ndim,JACKNUM,MOCK,MOCKFUNC,ERRLEN,HPNUM,STARTDX,ANUM;
	int ALGORITHM,PRIORMODEL,FILEFORMAT,SVDEXTEND,MYRANK,MPINODES,COMINT,KERNELTYPE;
	int CHECKSINGSUB, MAXITER, LBFGSAPPROX, FULLQ,PMULTIPLE, FULLERR;
	int SEARCHBASIS;
	double LDIFFTOL, LDIFFTOLPRECOND,PRIORPOWER;
	
	MYRANK=0;
	
	char config_path[600],lat_fname[600],fname[400],FILENAME[400],RESULTNAME[400];

	FILE* avgout;
	FILE* rho_output;
	FILE* data_output;
		
	gslgmp_vector* tau;		//# of points in imaginary time direction
	gslgmp_vector* omega;	//# of points in frequency space
    
	//*************************************************************************
	//	Read the command line parameters
	//*************************************************************************
	
	if (argc!=45) {fprintf(stderr, "ARGUMENTS GIVEN:%d must be %d \n",argc,45); exit(-1);}


        //nice(10);

	NX=atoi(argv[1]);
	XMAX=atoi(argv[2]);
	NT=atoi(argv[3]);
	TREADMAX=atoi(argv[4]);
	NTMIN=atoi(argv[5]);
	NTMAX=atoi(argv[6]);
	NOMEGA=atoi(argv[7]);
	WMIN=(double)atof(argv[8]);
	WMAX=(double)atof(argv[9]);
	PRIORPOWER=atof(argv[10]);
	MFAC=(double)atof(argv[11]);
	ERRNUM=atoi(argv[12]);
	NUMCONF=atoi(argv[13]);
	JACKNUM=atoi(argv[14]);
	FOFFSET=atoi(argv[15]);
	BETA=(double)atof(argv[16]);
	TWIDTH=NTMAX-NTMIN+1;
	sprintf(config_path,"%s",argv[17]);
	MOCK=atoi(argv[18]);
	MOCKNOISE=(double)atof(argv[19]);
	MOCKFUNC=atoi(argv[20]);
	HPSTART=(double)atof(argv[21]);
	HPEND=(double)atof(argv[22]);
	HPNUM=atoi(argv[23]);
	STARTDX=atoi(argv[24]);
	ASTART=(double)atof(argv[25]);
	AMAX=(double)atof(argv[26]);
	AMIN=(double)atof(argv[27]);
	ANUM=atoi(argv[28]);
	ALGORITHM=atoi(argv[29]);

	LDIFFTOL=(double)atof(argv[30]);
	LDIFFTOLPRECOND=(double)atof(argv[31]);
	
	MAXITER=atoi(argv[32]);
	LBFGSAPPROX=atoi(argv[33]);
	FULLQ=atoi(argv[34]);

	SVDEXTEND=atoi(argv[35]);
	PRIORMODEL=atoi(argv[36]);

	FILEFORMAT=atoi(argv[37]);
	sprintf(FILENAME,"%s",argv[38]);
	sprintf(RESULTNAME,"%s",argv[39]);
	KERNELTYPE=atoi(argv[40]);

	if(ALGORITHM>0) { SEARCHBASIS=KERNELTYPE;  }
	if(ALGORITHM<0) { SEARCHBASIS=-KERNELTYPE; ALGORITHM*=-1;}

	CHECKSINGSUB=atoi(argv[41]);
	
	FULLERR=atoi(argv[44]);

	//*************************************************************************
	//	SET THE GLOBAL PRECISION OF THIS RUN
	//*************************************************************************
	
	if(PRECISION_SET==0){
		PRECBITS=atoi(argv[42]);
		PMULTIPLE=atoi(argv[43]);
		Precision=PRECBITS;
		PRECISION_SET=1;
	}
	
	fprintf(stderr, "\n**********************************************************\n");
	fprintf(stderr, "***   GLOBAL PRECISION SET TO %3d BITS                ***\n",PRECBITS);
	fprintf(stderr, "**********************************************************\n");
	
	//*************************************************************************
	//	IF WE WANT TO RUN THE DEBUGIN ROUTINES DO IT HERE
	//*************************************************************************
	#if (DEBUGTST==1)
	TESTSUITE();
	#endif
	//*************************************************************************
	//	DEPENDING ON THE INPUT SET CONSISTENT PARAMETERS
	//*************************************************************************
	
	if(NTMAX>TREADMAX||NTMIN>TREADMAX) {fprintf(stderr,"TREADMAX is smaller than the measured data, exiting!\n"); exit(-1);}
	if((NOMEGA%ERRNUM)!=0 || ERRNUM>=NOMEGA) {fprintf(stderr,"ERRNUM must divide NOMEGA without rest!\n"); exit(-1);}

        if(JACKNUM>0) if(NUMCONF%JACKNUM!=0) { fprintf(stderr,"JACKNUM MUST DIVIDE NUMCONF WITHOUT REST!\n"); exit(-1);}

	ERRLEN=NOMEGA/ERRNUM;

	//*************************************************************************
	//	OUTPUT THE PARAMETERS TO FILE                   
	//*************************************************************************

		fprintf(stderr, "\n**********************************************************\n");
		fprintf(stderr, "***   PARAMETERS                                       ***\n");
		fprintf(stderr, "***     NX: %3d                                        ***\n",NX);
		fprintf(stderr, "***     XMAX: %3d                                      ***\n",XMAX);
		fprintf(stderr, "***     NT: %3d                                        ***\n",NT);
		fprintf(stderr, "***     TREADM: %3d                                    ***\n",TREADMAX);
		fprintf(stderr, "***     NTMIN: %3d                                     ***\n",NTMIN);
		fprintf(stderr, "***     NTMAX: %3d                                     ***\n",NTMAX);
		fprintf(stderr, "***     NOMEGA: %3d                                    ***\n",NOMEGA);
		fprintf(stderr, "***     WMIN: %3f                                      ***\n",WMIN);
		fprintf(stderr, "***     WMAX: %3f                                      ***\n",WMAX);
		fprintf(stderr, "***     PRIORP: %3f                                    ***\n",PRIORPOWER);
		fprintf(stderr, "***     MFAC: %3f                                      ***\n",MFAC);
		fprintf(stderr, "***     ERRNUM: %3d                                    ***\n",ERRNUM);
		fprintf(stderr, "***     NUMCONF: %3d                                   ***\n",NUMCONF);
		fprintf(stderr, "***     JACKNUM: %3d                                   ***\n",JACKNUM);
		fprintf(stderr, "***     FOFFS: %3d                                     ***\n",FOFFSET);
		fprintf(stderr, "***     BETA: %3f                                      ***\n",BETA);
		fprintf(stderr, "***     CONFIGP: %s                                    ***\n",config_path);
		fprintf(stderr, "***     MOCK: %d                                       ***\n",MOCK);
		fprintf(stderr, "***     MOCKN: %3f                                     ***\n",MOCKNOISE);
		fprintf(stderr, "***     MOCKF: %3d                                     ***\n",MOCKFUNC);
		fprintf(stderr, "***     HPS: %3f                                       ***\n",HPSTART);
		fprintf(stderr, "***     HPE: %3f                                       ***\n",HPEND);
		fprintf(stderr, "***     HPNUM: %3d                                     ***\n",HPNUM);
		fprintf(stderr, "***     STARTX: %3d                                    ***\n",STARTDX);
		fprintf(stderr, "***     ASTART: %3f                                    ***\n",ASTART);
		fprintf(stderr, "***     AMAX: %3f                                      ***\n",AMAX);
		fprintf(stderr, "***     AMIN: %3f                                      ***\n",AMIN);
		fprintf(stderr, "***     ANUM: %3d                                      ***\n",ANUM);
		fprintf(stderr, "***     ALGOR: %3d                                     ***\n",ALGORITHM);
		fprintf(stderr, "***     TOLER: %3f                                     ***\n",LDIFFTOL);
		fprintf(stderr, "***     MAXIT: %3d                                     ***\n",MAXITER);
		fprintf(stderr, "***     LBFGS: %3d                                     ***\n",LBFGSAPPROX );
		fprintf(stderr, "***     FULLQ: %3d                                     ***\n",FULLQ );
		fprintf(stderr, "***     SVDEXT: %3d                                    ***\n",SVDEXTEND);
		fprintf(stderr, "***     PRIORM: %3d                                    ***\n",PRIORMODEL);
		fprintf(stderr, "***     FFORM: %3d                                     ***\n",FILEFORMAT);
		fprintf(stderr, "***     FNAME: %s                                      ***\n",FILENAME);
		fprintf(stderr, "***     RPATH: %s                                      ***\n",RESULTNAME);
		fprintf(stderr, "***     KTYPE: %3d                                     ***\n",KERNELTYPE);
		fprintf(stderr, "***     CSUB: %3d                                      ***\n",CHECKSINGSUB);
		fprintf(stderr, "***     PBITS: %3d                                     ***\n",PRECBITS);
		fprintf(stderr, "***     PMULT: %3d                                     ***\n",PMULTIPLE);
		fprintf(stderr, "***     FULER: %3d                                     ***\n",FULLERR);
		fprintf(stderr, "***                                                    ***\n");
		fprintf(stderr, "**********************************************************\n");

	
	    FILE* outputfile;


		sprintf(fname, "%s%s", RESULTNAME, "output.log");
		outputfile=fopen(fname,"w");

		sprintf(fname, "mkdir %s", RESULTNAME);
		if(outputfile==0){ system (fname);}else{fclose(outputfile);}
		
		sprintf(fname, "%s%s", RESULTNAME, "output.log");
		outputfile=fopen(fname,"w");


		sprintf(fname, "%s%s", RESULTNAME, "parameters.log");
		outputfile=fopen(fname,"w");
		
		
		fprintf(outputfile, "\n**********************************************************\n");
		fprintf(outputfile, "***   PARAMETERS                                       ***\n");
		fprintf(outputfile, "***     NX: %3d                                        ***\n",NX);
		fprintf(outputfile, "***     XMAX: %3d                                      ***\n",XMAX);
		fprintf(outputfile, "***     NT: %3d                                        ***\n",NT);
		fprintf(outputfile, "***     TREADM: %3d                                    ***\n",TREADMAX);
		fprintf(outputfile, "***     NTMIN: %3d                                     ***\n",NTMIN);
		fprintf(outputfile, "***     NTMAX: %3d                                     ***\n",NTMAX);
		fprintf(outputfile, "***     NOMEGA: %3d                                    ***\n",NOMEGA);
		fprintf(outputfile, "***     WMIN: %3f                                      ***\n",WMIN);
		fprintf(outputfile, "***     WMAX: %3f                                      ***\n",WMAX);
		fprintf(outputfile, "***     PRIORP: %3f                                    ***\n",PRIORPOWER);
		fprintf(outputfile, "***     MFAC: %3f                                      ***\n",MFAC);
		fprintf(outputfile, "***     ERRNUM: %3d                                    ***\n",ERRNUM);
		fprintf(outputfile, "***     NUMCONF: %3d                                   ***\n",NUMCONF);
		fprintf(outputfile, "***     JACKNUM: %3d                                   ***\n",JACKNUM);
		fprintf(outputfile, "***     FOFFS: %3d                                     ***\n",FOFFSET);
		fprintf(outputfile, "***     BETA: %3f                                      ***\n",BETA);
		fprintf(outputfile, "***     CONFIGP: %s                                    ***\n",config_path);
		fprintf(outputfile, "***     MOCK: %d                                       ***\n",MOCK);
		fprintf(outputfile, "***     MOCKN: %3f                                     ***\n",MOCKNOISE);
		fprintf(outputfile, "***     MOCKF: %3d                                     ***\n",MOCKFUNC);
		fprintf(outputfile, "***     HPS: %3f                                       ***\n",HPSTART);
		fprintf(outputfile, "***     HPE: %3f                                       ***\n",HPEND);
		fprintf(outputfile, "***     HPNUM: %3d                                     ***\n",HPNUM);
		fprintf(outputfile, "***     STARTX: %3d                                    ***\n",STARTDX);
		fprintf(outputfile, "***     ASTART: %3f                                    ***\n",ASTART);
		fprintf(outputfile, "***     AMAX: %3f                                      ***\n",AMAX);
		fprintf(outputfile, "***     AMIN: %3f                                      ***\n",AMIN);
		fprintf(outputfile, "***     ANUM: %3d                                      ***\n",ANUM);
		fprintf(outputfile, "***     ALGOR: %3d                                     ***\n",ALGORITHM);
		fprintf(outputfile, "***     TOLER: %3f                                     ***\n",LDIFFTOL);
		fprintf(outputfile, "***     MAXIT: %3d                                     ***\n",MAXITER);
		fprintf(outputfile, "***     LBFGS: %3d                                     ***\n",LBFGSAPPROX );
		fprintf(outputfile, "***     FULLQ: %3d                                     ***\n",FULLQ );
		fprintf(outputfile, "***     SVDEXT: %3d                                    ***\n",SVDEXTEND);
		fprintf(outputfile, "***     PRIORM: %3d                                    ***\n",PRIORMODEL);
		fprintf(outputfile, "***     FFORM: %3d                                     ***\n",FILEFORMAT);
		fprintf(outputfile, "***     FNAME: %s                                      ***\n",FILENAME);
		fprintf(outputfile, "***     RPATH: %s                                      ***\n",RESULTNAME);
		fprintf(outputfile, "***     KTYPE: %3d                                     ***\n",KERNELTYPE);
		fprintf(outputfile, "***     CSUB: %3d                                      ***\n",CHECKSINGSUB);
		fprintf(outputfile, "***     PBITS: %3d                                     ***\n",PRECBITS);
		fprintf(outputfile, "***     PMULT: %3d                                     ***\n",PMULTIPLE);
		fprintf(outputfile, "***     FULER: %3d                                     ***\n",FULLERR);
		fprintf(outputfile, "***                                                    ***\n");
		fprintf(outputfile, "**********************************************************\n");
		

		fclose(outputfile);
		
	
	//*************************************************************************
	//	INITIALIZE THE OUTPUT FILES
	//*************************************************************************

		sprintf(fname, "%s%s", RESULTNAME, "spec_rec.dat");
		outputfile=fopen(fname,"w");
		fclose(outputfile);

		sprintf(fname, "%s%s", RESULTNAME, "spec_recdat.dat");
		outputfile=fopen(fname,"w");
		fclose(outputfile);

		sprintf(fname, "%s%s", RESULTNAME, "JACK_spec_rec.dat");
		outputfile=fopen(fname,"w");
		fclose(outputfile);

		sprintf(fname, "%s%s", RESULTNAME, "JACK_spec_recdat.dat");
		outputfile=fopen(fname,"w");
		fclose(outputfile);

		sprintf(fname, "%s%s", RESULTNAME, "MEM_DCDXDATA.dat");
		outputfile=fopen(fname,"w");
		fclose(outputfile);

		sprintf(fname, "%s%s", RESULTNAME, "Probabilities.dat");
		outputfile=fopen(fname,"w");
		fclose(outputfile);

		sprintf(fname, "%s%s", RESULTNAME, "Prior.dat");
		outputfile=fopen(fname,"w");
		fclose(outputfile);

		sprintf(fname, "%s%s", RESULTNAME, "MEM_spec_rec_alpha.dat");
		outputfile=fopen(fname,"w");
		fclose(outputfile);

		sprintf(fname, "%s%s", RESULTNAME, "Intervals.dat");
		outputfile=fopen(fname,"w");
		fclose(outputfile);
		
		sprintf(fname, "%s%s", RESULTNAME, "Projvals.dat");
		outputfile=fopen(fname,"w");
		fclose(outputfile);
	
	//*************************************************************************
	//	Initialize the TEMPORARY VARIABLES
	//*************************************************************************

	mpfr_t ptmp1,ptmp2,ptmp3,ptmp4,ptmp5,ptmp6;	// Temporary high precision variables

	mpfr_init2(ptmp1,PRECBITS);
	mpfr_init2(ptmp2,PRECBITS);
	mpfr_init2(ptmp3,PRECBITS);
	mpfr_init2(ptmp4,PRECBITS);
	mpfr_init2(ptmp5,PRECBITS);
	mpfr_init2(ptmp6,PRECBITS);

	//*************************************************************************
	//	Declare the basic quantities
	//*************************************************************************

	gslgmp_matrix* Kernel;		//The original Kernel of the system
	gslgmp_matrix* DCKernel;	//The Kernel after diagonalizing correlations

	gslgmp_matrix *U,*V;		//To store the full SVD
	gslgmp_vector *Xi;		//To store the SVD singular values
	gslgmp_matrix *Us;		//To store the singular subspace part of U
	int NSING;
	int NSINGEXT;

	gslgmp_vector* DXData;		//Imaginary time dataset for current DX slice
	gslgmp_vector* DCDXData;	//Imaginary time dataset for current DX slice after decorrelation

	gslgmp_vector* DXDA;		//Imaginary time dataset from the reconstructed spectral function values
	gslgmp_vector* DXA;		//The reconstructed spectral function A for the current DX slice
	gslgmp_vector* DXb;		//The reconstructed spectral function A for the current DX slice
	gslgmp_vector* DXAdw;		//The reconstructed spectral function A for the current DX slice
	gslgmp_vector* DXAA;		//The reconstructed spectral function A for the current DX slice
	gslgmp_vector* DA;		//The reconstructed spectral function A for the current DX slice
	gslgmp_vector* DAWCOR;		//The reconstructed spectral function A for the current DX slice
	
	gslgmp_vector* A;		//a vector to contain the current Spectral reconstruction
	gslgmp_vector* Adw;		//a vector to contain the current Spectral reconstruction with the domega terms multiplied
	gslgmp_vector* a;		//a vector to construct A
	gslgmp_vector* blbfgs;		//a vector to construct A
	gslgmp_vector* blbfgsSVD;	//a vector to construct A
	gslgmp_vector* blm;		//a vector to construct A
	gslgmp_vector* b;		//a vector to construct A
	gslgmp_vector* blbfgsback;	//a vector to construct A
	gslgmp_vector* blmback;		//a vector to construct A

	gslgmp_vector* Xis;		//To store the singular subspace part of Xi
	gslgmp_matrix* ddLddDA;		//Inverse of Correlation Eigenvalues
	gslgmp_vector* invSigm;		//Inverse of the Correlation Eigenvallues as vector

	gslgmp_matrix* Vs;		//To store the singular subspace part of V
	gslgmp_matrix* XisM;		//To store the singular subspace part of Xi as diagonal matrix
	gslgmp_matrix* UsExt;

	gslgmp_vector* mprior;		//Prior model
	gslgmp_vector* mpriorback;	//Prior model
	gslgmp_vector* mpriordw;	//Prior model

	gslgmp_matrix* Corr;		//Correlation Matrix
	gslgmp_matrix* CInv;		//Correlation Matrix Inverse
	gslgmp_matrix* CEvec;		//Eigenvectors of the Correlation Matrix
	gslgmp_vector* CEval;		//Eigenvalues of the Correlation Matrix

	gslgmp_matrix* M2,*T2;		//Matrices used in the Calculation of the Probability for alpha

	gslgmp_matrix* Lambda;		//Correlation Matrix
	gslgmp_matrix* LEvec;		//Eigenvectors of the Correlation Matrix
	gslgmp_vector* LEval;		//Eigenvalues of the Correlation Matrix
	gslgmp_vector* LOGPROB;		//Eigenvalues of the Correlation Matrix

	gslgmp_matrix* ErrMatLoc;	//Local Error Matrix
	gslgmp_matrix* ErrMatInv;	//Invverse of Local Error Matrix


	double* DATA;
	int acnt;

	//*************************************************************************
	//	Allocate the basic quantities
	//*************************************************************************

	Kernel=gslgmp_matrix_calloc(TWIDTH,NOMEGA);
	tau=gslgmp_vector_calloc(TWIDTH);
	omega=gslgmp_vector_calloc(NOMEGA);
	mprior=gslgmp_vector_calloc(NOMEGA);
	mpriorback=gslgmp_vector_calloc(NOMEGA);
	mpriordw=gslgmp_vector_calloc(NOMEGA);

	//*************************************************************************
	//	Allocate the arrays for the Kernel^t SVD decomposition
	//*************************************************************************

	U=gslgmp_matrix_calloc(NOMEGA,TWIDTH);
	V=gslgmp_matrix_calloc(TWIDTH,TWIDTH);
	Xi=gslgmp_vector_calloc(TWIDTH);


	//*************************************************************************
	//	Allocate the Lattice data and Average arrays
	//*************************************************************************

	DATA=(double*)malloc(NUMCONF*NT*sizeof(double));

	//*****************************************************************
	//	Initialize the Correlations Matrices and related quantities
	//	as well as the arrays for the decorrelated Data and Kernel
	//*****************************************************************

	CInv=gslgmp_matrix_calloc(TWIDTH,TWIDTH);

	//*****************************************************************
	//	Initialize the DXDA, DXDATA, DADX and a vector
	//*****************************************************************

	DCKernel=gslgmp_matrix_calloc(TWIDTH,NOMEGA);
	DXData=gslgmp_vector_calloc(TWIDTH);
	DXDA=gslgmp_vector_calloc(TWIDTH);
	DXA=gslgmp_vector_calloc(NOMEGA);
	DXb=gslgmp_vector_calloc(TWIDTH);

	gslgmp_vector* g;
	gslgmp_vector* db;
	gslgmp_vector* bold;
	gslgmp_vector* eta;
	gslgmp_vector* tst;
	gslgmp_vector* dLdDA;

	gslgmp_matrix* SV;
	gslgmp_matrix* T;
	gslgmp_matrix* M;
	gslgmp_matrix* Chi;
	gslgmp_matrix* J;

	
	//*****************************************************************
	//	Print to screen which Minimizer is used        	
	//*****************************************************************


	if (ALGORITHM==10){
		
		fprintf(stderr, "\n**********************************************************\n");
		fprintf(stderr, "***   MEM WITH LM    MINIMIZATION USING THE ADDITIONAL ***\n");
		fprintf(stderr, "***   SVD PRESCRIPTION  AND A(w)=m(w)exp(a(w))         ***\n");
		fprintf(stderr, "***   WITH EXTENDED SEARCH SPACE NEXT: %3d             ***\n", TWIDTH+SVDEXTEND);
		fprintf(stderr, "***   WITH LBFGS PRECONDITIONER                        ***\n");
		fprintf(stderr, "**********************************************************\n");
		
	}else if (ALGORITHM==6){
		
		fprintf(stderr, "\n**********************************************************\n");
		fprintf(stderr, "***   MEM WITH LM    MINIMIZATION USING THE ADDITIONAL ***\n");
		fprintf(stderr, "***   SVD PRESCRIPTION  AND A(w)=m(w)exp(a(w))         ***\n");
		fprintf(stderr, "***   WITH LBFGS PRECONDITIONER                        ***\n");
		fprintf(stderr, "**********************************************************\n");
		
	}else if(ALGORITHM==5){
		
		fprintf(stderr, "\n**********************************************************\n");
		fprintf(stderr, "***   MEM WITH LBFGS MINIMIZATION USING THE ADDITIONAL ***\n");
		fprintf(stderr, "***   SVD PRESCRIPTION  AND A(w)=m(w)exp(a(w))         ***\n");
		fprintf(stderr, "***   WITH THE HESSIAN APPROXIMATED BY %3d STEPS       ***\n",LBFGSAPPROX);
		fprintf(stderr, "**********************************************************\n");


	}else if (ALGORITHM==2){
		
		fprintf(stderr, "\n**********************************************************\n");
		fprintf(stderr, "***   MEM WITH LBFGS MINIMIZATION USING THE NON        ***\n");
		fprintf(stderr, "***   SVD PRESCRIPTION IN NOMEGA D.O.F.                ***\n");
		fprintf(stderr, "***   WITH THE HESSIAN APPROXIMATED BY %3d STEPS       ***\n",LBFGSAPPROX);
		fprintf(stderr, "**********************************************************\n");
		
	}else{
		
		fprintf(stderr, "\n***************************************************************************\n");
		fprintf(stderr, "***   PLEASE CHOOSE ONE OF THE ALGORITHMS BELOW:             		***\n");
		fprintf(stderr, "***                                                      		***\n");
		fprintf(stderr, "***  10. LBFGS PRECONDITIONED LM WITH SVD IN EXTENDED SEARCH SPACE     ***\n");
		fprintf(stderr, "***   6. LBFGS PRECONDITIONED LM WITH SVD IN NSING<NTAU D.O.F.		***\n");
		fprintf(stderr, "***   5. LBFGS WITH SVD PRESCRIPTION IN NSING<NTAU D.O.F.		***\n");
		fprintf(stderr, "***   2. LBFGS IN NOMEGA D.O.F.					***\n");
		fprintf(stderr, "******************************************************************************\n");
		
		exit(-1);

	}


	//*****************************************************************
	//	INITIALIZATION OF THE OMEGA INTERVAL
	//
	//	IF HPNUM=0 WE USE EQUIDISTANT SPACINGS, HOWEVER IF A
	//	FINE INTERVAL IS CHOSEN, WE HAVE TWO DIFFERENT SPACINGS
	//	FOR OMEGA
	//
	//*****************************************************************

	int OMEGAHPMIN=0;
	int OMEGAHPMAX=0;

	double wtmp;
	double step;

	if(HPNUM==0){
    
		fprintf(stderr, "\n**********************************************************\n");
		fprintf(stderr, "***   INITIALIZING THE FREQUENCY RANGE                ***\n");
		fprintf(stderr, "**********************************************************\n");
	
		#if (WITHGUI==1) 
		    wxMutexGuiEnter();
		    *TInfo->threadtxt << wxT("Initializing the frequency range\n") ;
		    wxMutexGuiLeave();
		#endif
		
		for(int dw=0;dw<NOMEGA;dw++){
	
			wtmp=WMIN+(WMAX-WMIN)*((double)dw)/((double)NOMEGA-1.);
			gslgmp_vector_set_d(omega,dw,(wtmp));
		}

		OMEGAHPMAX=-1;

	}else{

		fprintf(stderr, "\n**********************************************************\n");
		fprintf(stderr, "***   INITIALIZING THE FREQUENCY RANGE WITH A HIGH   ***\n");
		fprintf(stderr, "***   RESOLUTION AREA FROM [%.2f,%.2f] WITH %4d     ***\n",HPSTART,HPEND,HPNUM);
		fprintf(stderr, "**********************************************************\n");

		double nstep=((WMAX-WMIN)-(HPEND-HPSTART))/((double)(NOMEGA-1.-HPNUM));
		double hpstep=((HPEND-HPSTART))/((double)(HPNUM));

		wtmp=WMIN;
		gslgmp_vector_set_d(omega,0,(wtmp));
		
		for(int dw=1;dw<NOMEGA;dw++){

			if(wtmp<HPSTART||wtmp>=HPEND) { step=nstep; if(wtmp>=HPEND&&OMEGAHPMAX==0){OMEGAHPMAX=dw-1;} }
						      else
						      {if(OMEGAHPMIN==0) {OMEGAHPMIN=dw-1 ;} step=hpstep;}
			
			wtmp=wtmp+step;

			gslgmp_vector_set_d(omega,dw,(wtmp));

		}
		
		fprintf(stderr, "\n**********************************************************\n");
		fprintf(stderr, "***   INITIALIZED  THE FREQUENCY RANGE WITH A HIGH   ***\n");
		fprintf(stderr, "***   RESOLUTION AREA FROM [%4d ,%4d ] WITH %4d     ***\n",OMEGAHPMIN,OMEGAHPMAX,HPNUM);
		fprintf(stderr, "**********************************************************\n");
		
	}
	
	
	//*****************************************************************
	//	Initialize the tau vector
	//*****************************************************************

	fprintf(stderr, "\n**********************************************************\n");
	fprintf(stderr, "***   INITIALIZING THE TAU TIME  RANGE                ***\n");
	fprintf(stderr, "**********************************************************\n");
		
  	for(int dt=NTMIN;dt<=NTMAX;dt++){
  
  		gslgmp_vector_set_d(tau,dt-NTMIN,BETA*((double)dt)/((double)NT));
  	}

	//*****************************************************************
	//	GENERATE THE DOUBLE VARIABLE VERSION OF TAU AND OMEGA
	//*****************************************************************

	double* domega;
	double* dtau;

	domega=(double*)malloc(sizeof(double)*NOMEGA);
	dtau=(double*)malloc(sizeof(double)*TWIDTH);

	for(int dw=0;dw<NOMEGA;dw++) domega[dw]=gslgmp_vector_get_d(omega,dw);
	for(int dt=0;dt<TWIDTH;dt++) dtau[dt]=gslgmp_vector_get_d(tau,dt);

	//*****************************************************************
	//	FIND THE INTEGER POSITION OF THE FIRST NON NEGATIVE  
	//	OMEGA
	//*****************************************************************

	int POSNW;

	for(int dw=0;dw<NOMEGA;dw++) if(domega[dw]>0.){ POSNW=dw; break;} 

	if(WMIN<0){
	fprintf(stderr, "\n**********************************************************\n");
	fprintf(stderr, "***   THE FIRST POSITIVE FREQUENCY IN THE OMEGA ARRAY  ***\n");
	fprintf(stderr, "***   IS AT POSITION %4d = %e                      ***\n",POSNW,domega[POSNW]);
	fprintf(stderr, "**********************************************************\n");
	}

	//*****************************************************************
	//	GENERATE THE KERNEL ACCORDING TO OUR CHOICE OF KERNEL 
	//	FUNCTION:
	//
	//	0. The zero temperature Kernel for the Spectral functions
	//	1. The naive finite temperature Kernel
	//
	//*****************************************************************

	fprintf(stderr, "\n**********************************************************\n");
	fprintf(stderr, "***   INITIALIZING THE INTEGRAL KERNEL                ***\n");
	fprintf(stderr, "**********************************************************\n");
	#if (WITHGUI==1) 
	    wxMutexGuiEnter();
	    *TInfo->threadtxt << wxT("Initializing the Kernel\n") ;
	    wxMutexGuiLeave();
	#endif

	if(KERNELTYPE==1){
	    
	    //*************************************************************
	    //		Zero temperature Kernel
	    //		K(w_i,t_j)=Exp(-w_i*t_j)
	    //*************************************************************
	    
	fprintf(stderr, "\n**********************************************************\n");
	fprintf(stderr, "***   ZERO TEMPERTURE KERNEL K(w_i,t_j)=Exp(-w_i*t_j) ***\n");
	fprintf(stderr, "**********************************************************\n");

	    mpfr_t ktmp;
	    mpfr_init2(ktmp,PRECBITS);

	    for(int dt=0;dt<TWIDTH;dt++){
	    for(int dw=0;dw<NOMEGA;dw++){

		    gslgmp_vector_get(tau,dt,ptmp1);
		    gslgmp_vector_get(omega,dw,ptmp2);
		    mpfr_mul(ptmp3,ptmp1,ptmp2,GMP_RNDN);
		    mpfr_neg(ptmp1,ptmp3,GMP_RNDN);
		    mpfr_exp (ktmp, ptmp1, GMP_RNDN);

		    gslgmp_matrix_set(Kernel,dt,dw,ktmp);


	    }}
	
	} else if(KERNELTYPE==2){
	
	    //*************************************************************
	    //		Naive finite temperature Kernel(Half BETA)
	    //		K(w_i,t_j)= Cosh[w(t-0.5\beta)]/Sinh[w*0.5\beta]
	    //*************************************************************
	fprintf(stderr, "\n**********************************************************\n");
	fprintf(stderr, "***   Naive finite temperature Kernel                 ***\n");
	fprintf(stderr, "***   K(w_i,t_j)= Cosh[w(t-0.5\beta)]/Sinh[w*0.5\beta]***\n");
	fprintf(stderr, "**********************************************************\n");
	
	    mpfr_t ktmp;
	    mpfr_init2(ktmp,PRECBITS);

	    for(int dt=0;dt<TWIDTH;dt++){
	    for(int dw=0;dw<NOMEGA;dw++){

		    gslgmp_vector_get(tau,dt,ptmp1);
		    gslgmp_vector_get(omega,dw,ptmp2);
		    
		    mpfr_sub_d(ptmp3, ptmp1,0.5*BETA,GMP_RNDN);
		    mpfr_mul(ptmp3,ptmp3,ptmp2,GMP_RNDN);
		    mpfr_cosh(ptmp1,ptmp3,GMP_RNDN);
		    
		    mpfr_mul_d(ptmp3,ptmp2,0.5*BETA,GMP_RNDN);
		    mpfr_sinh(ptmp2,ptmp3,GMP_RNDN);

		    mpfr_div(ptmp3,ptmp1,ptmp2,GMP_RNDN);
		    
		    gslgmp_matrix_set(Kernel,dt,dw,ptmp3);
		    

	    }}
	
	} else if(KERNELTYPE==3){
	
	    //*************************************************************
	    //		Regularized finite temperature Kernel
	    //		K(w_i,t_j)= Tanh[omega\beta/2]*Cosh[w(t-0.5\beta)]/Sinh[w*0.5\beta]
	    //                    = Cosh[w(t-0.5\beta)]/Cosh[\omega\beta/2]
	    //*************************************************************
	fprintf(stderr, "\n**********************************************************\n");
	fprintf(stderr, "***   Regularized finite Temperature Kernel           ***\n");
	fprintf(stderr, "***   K(w_i,t_j)= Cosh[w(t-0.5\beta)]/Cosh[w*0.5\beta]***\n"); 
	fprintf(stderr, "**********************************************************\n");
	
	    mpfr_t ktmp;
	    mpfr_init2(ktmp,PRECBITS);

	    for(int dt=0;dt<TWIDTH;dt++){
	    for(int dw=0;dw<NOMEGA;dw++){

		    gslgmp_vector_get(tau,dt,ptmp1);
		    gslgmp_vector_get(omega,dw,ptmp2);
		    
		    mpfr_sub_d(ptmp3, ptmp1,0.5*BETA,GMP_RNDN);
		    mpfr_mul(ptmp3,ptmp3,ptmp2,GMP_RNDN);
		    mpfr_cosh(ptmp1,ptmp3,GMP_RNDN);
		    
		    mpfr_mul_d(ptmp3,ptmp2,0.5*BETA,GMP_RNDN);
		    mpfr_cosh(ptmp2,ptmp3,GMP_RNDN);

		    mpfr_div(ptmp3,ptmp1,ptmp2,GMP_RNDN);
		    
		    gslgmp_matrix_set(Kernel,dt,dw,ptmp3);
		    

	    }}
	
	} else {
	    
	    fprintf(stderr, "\n**********************************************************\n");
	    fprintf(stderr, "***   PLEASE CHOOSE A KERNEL THAT IS IMPLEMENTED  KT=1 ***\n");
	    fprintf(stderr, "**********************************************************\n");
	    
	    exit(-1);
	    
	}
	
	//*****************************************************************
	//	Print the tau values to the screen.
	//*****************************************************************
	
	gmpPVec(tau);

	//*************************************************************************
	//	START TO GO THROUGH THE INDIVIDUAL MEASUREMENTS SETS
	//	SINCE THESE WERE ORIGINALLY MEASUREMENTS FOR DIFFERENT X THE
	//	COUNTER IS STILL CALLED DX
	//*************************************************************************
	
	for(int DX=STARTDX;DX<=abs(XMAX);DX++){

	    fprintf(stderr, "\n**********************************************************\n");
	    fprintf(stderr, "***   NODE %3d STARTING WITH DX:%3d                  ***\n",MYRANK,DX);
	    fprintf(stderr, "**********************************************************\n");

	    #if (WITHGUI==1) 
   	        wxMutexGuiEnter();
		*TInfo->threadtxt << wxT("Starting with DX: [") << DX << wxT("] \n");
	        wxMutexGuiLeave();
	    #endif
	    
	    
	    //*************************************************************************
	    //	SET THE PRIOR MODEL FOR THIS MEASUREMENT SET:
	    //
	    //	1.   	m=mfac* w^PRIORPOWER
	    //	2.   	m=mfac*1/w
	    //	3. 	m=mfac
	    //*************************************************************************

	    if (PRIORMODEL==1){

		    fprintf(stderr, "\n**********************************************************\n");
		    fprintf(stderr, "***   USING THE PRIOR (w-w0)^(%3f)                       ***\n",PRIORPOWER);
		    fprintf(stderr, "**********************************************************\n");

		    for(int dw=0;dw<NOMEGA;dw++){

			    mpfr_set(ptmp1,omega->data[dw],GMP_RNDN);
			    mpfr_set(ptmp2,omega->data[0],GMP_RNDN);
			    mpfr_sub(ptmp3,ptmp1,ptmp2,GMP_RNDN);

			    mpfr_log(ptmp1,ptmp3,GMP_RNDN);
			    mpfr_mul_d(ptmp1,ptmp1,PRIORPOWER,GMP_RNDN);
			    mpfr_exp(ptmp1,ptmp1,GMP_RNDN);

			    mpfr_set_d(ptmp2,MFAC,GMP_RNDN);
			    mpfr_add(mprior->data[dw],ptmp1,ptmp2,GMP_RNDN);
			    
		    }
		    
	    }else if (PRIORMODEL==2){
		    
		    fprintf(stderr, "\n**********************************************************\n");
		    fprintf(stderr, "***   USING THE PRIOR (w-w0+1)^(%3f) WITH m(w_min)=m_0   ***\n",PRIORPOWER);
		    fprintf(stderr, "**********************************************************\n");
		    
		    for(int dw=0;dw<NOMEGA;dw++){

			    mpfr_set(ptmp1,omega->data[dw],GMP_RNDN);
			    mpfr_set(ptmp2,omega->data[0],GMP_RNDN);
			    mpfr_sub(ptmp3,ptmp1,ptmp2,GMP_RNDN);
			    mpfr_add_ui(ptmp3,ptmp3,1,GMP_RNDN);

			    mpfr_log(ptmp1,ptmp3,GMP_RNDN);
			    mpfr_mul_d(ptmp1,ptmp1,PRIORPOWER,GMP_RNDN);
			    mpfr_exp(ptmp1,ptmp1,GMP_RNDN);
			    
			    mpfr_set_d(ptmp2,MFAC,GMP_RNDN);
			    mpfr_mul(mprior->data[dw],ptmp1,ptmp2,GMP_RNDN);
		    }
		    
	    }else if (PRIORMODEL==3){
		    
		    fprintf(stderr, "\n**********************************************************\n");
		    fprintf(stderr, "***   USING THE CONSTANT PRIOR                         ***\n");
		    fprintf(stderr, "**********************************************************\n");
		    
		    for(int dw=0;dw<NOMEGA;dw++){
			    
			    mpfr_set_d(mprior->data[dw],MFAC,GMP_RNDN);
		    }
		    
	    }else if (PRIORMODEL==5){
		    
		    fprintf(stderr, "\n**********************************************************\n");
		    fprintf(stderr, "***   USING THE EXP EXP PRIOR                          ***\n");
		    fprintf(stderr, "**********************************************************\n");
		    
		    for(int dw=0;dw<NOMEGA;dw++){
			    
			    if(domega[dw]<0.){ 
			    mpfr_set(ptmp1,omega->data[dw],GMP_RNDN);
			    mpfr_mul_d(ptmp1,ptmp1,25,GMP_RNDN);
			    mpfr_exp(mprior->data[dw],ptmp1,GMP_RNDN);
			    }
			    
			    if(domega[dw]>=0.){ 
			    mpfr_set(ptmp1,omega->data[dw],GMP_RNDN);
			    mpfr_mul_d(ptmp1,ptmp1,-2,GMP_RNDN);
			    mpfr_exp(mprior->data[dw],ptmp1,GMP_RNDN);
			    }

			    
		}
		    
	    }else if (PRIORMODEL==4){
		    
		    fprintf(stderr, "\n**********************************************************\n");
		    fprintf(stderr, "***   USING A PREVIOUS RESULT AS PRIOR:                ***\n");
		    fprintf(stderr, "**********************************************************\n");
		    
		    FILE* INPRIOR;
		    INPRIOR=fopen("prior.dat","r");
		    
		    if(INPRIOR==0){ fprintf(stderr, "Cannot find prior file: prior.dat\n"); exit(-1);}
		    
		    for(int dw=0;dw<NOMEGA;dw++){
			    fprintf(stderr, "TEST BEFORE READ %d\n",dw);
			    double v1,v2,v3,v4,v5,v6,v7,v8,v9, v10,v11,v12;
			    int i1,i2,i3,i4,i5,i6;
			    
			    fscanf(INPRIOR,"%d %d %lf %lf %lf %lf %lf %d %d %lf %lf %lf %lf %lf %lf %d %d\n", &i1,&i2,&v1, &v2,&v3,&v4,&v5,&i3,&i4,&v6, &v7, &v8, &v9,&v10,&v11,&i5, &i6);
				    
			    if((ceil(1000.*v1))!=ceil(1000.*domega[dw])) { fprintf(stderr, "Frequency axis at [%d] %e != %e does not the same in prior and current run\n",dw,domega[dw],v1); exit(-1);}
			   
			    mpfr_set_d(mprior->data[dw],v4,GMP_RNDN);
			    mpfr_add_d(mprior->data[dw],mprior->data[dw],1e-40,GMP_RNDN);
		    }
		    
		    fclose(INPRIOR);
		   
	    }

	    //*********************************************************
	    //	ADD THE NECESSARY dw FACTORS FOR THE INTEGRAL
	    //*********************************************************

	    append_dw(mpriordw,mprior,omega,NOMEGA,HPNUM,OMEGAHPMIN,OMEGAHPMAX);

	    
	    if(FILEFORMAT==6&&NTMIN==0){
		    		
		mpfr_set_ui(ptmp1,0,GMP_RNDN);
		
		for(int w=0;w<NOMEGA;w++) mpfr_add(ptmp1,ptmp1,mpriordw->data[w],GMP_RNDN);		

		for(int dw=0;dw<NOMEGA;dw++){
			mpfr_div(mprior->data[dw], mprior->data[dw], ptmp1, GMP_RNDN);
			mpfr_div(mpriordw->data[dw], mpriordw->data[dw], ptmp1, GMP_RNDN);			
		}

		mpfr_set_ui(ptmp2,0,GMP_RNDN);
		
		for(int w=0;w<NOMEGA;w++) mpfr_add(ptmp2,ptmp2,mpriordw->data[w],GMP_RNDN);
		
		fprintf(stderr, "\n**********************************************************\n");
		fprintf(stderr, "***   PRIOR NORMALIZATION SET FOR FILEFORMAT 6         ***\n");
		fprintf(stderr, "***   INT OVER PRIOR1: %5e                  ***\n",mpfr_get_d(ptmp1,GMP_RNDN));
		fprintf(stderr, "***   INT OVER PRIOR2: %5e                  ***\n",mpfr_get_d(ptmp2,GMP_RNDN));
		fprintf(stderr, "**********************************************************\n");
		    
	    }
	    
	    
	    for(int dw=0;dw<NOMEGA;dw++){

		    mpfr_set(mpriorback->data[dw],mprior->data[dw],GMP_RNDN);

	    }
	    
		//*********************************************************************
		//	FOR EACH MEASUREMENT SET WE CAN PERFORM A JACKNIFE ANALYSIS.
		//	THIS RESAMPLING APPROACH PERFORMS THE MEM ANALYSIS FOR ALL
		//	MEASUREMENTS IN THE CURRENT SET WITH ONE BLOCK OF VALUES REMOVED.
		//  THE LENGTH OF EACH BLOCK DEPENDS ON THE NUMBER OF BLOCKS SPECIFIED
		//	IN THE COMMAND LINE ARGUMENTS: JACKNUM.
		//
		//	IN THE ZEROTHS REPETITION ALL MEASUREMENTS ARE USED IN THE MEM.
		//
		//*********************************************************************
	    
		int CONFL,CONFU; //These variables specify the start and finish of the Jacknife block 
	    
		double** FINALAJACK;	// These arrays contain the final Jacknifed MEM reconstruction
		double** FINALERRJACK;  // These arrays contain the final Jacknife error


		//*************************************************************************
		//	JACKNIFE ANALYSIS: 0th 		RUN WE USE ALL AVAILABLE DATA
		//			   1st-JACKNUM  RUN USES ALL DATA MINUS ONE BLOCK
		//			                SPECIFIED BY CONFL-CONFU
		//*************************************************************************
	
		if(JACKNUM!=0){
			FINALAJACK= (double**)malloc(sizeof(double*)*(JACKNUM));
			for(int i=0;i<JACKNUM+1;i++) FINALAJACK[i]=(double*)malloc(sizeof(double)*NOMEGA);
			FINALERRJACK= (double**)malloc(sizeof(double*)*(JACKNUM));
			for(int i=0;i<JACKNUM+1;i++) FINALERRJACK[i]=(double*)malloc(sizeof(double)*NOMEGA);
		}		
		
		//*************************************************************************
		//	THE JACKNIFE LOOP BEGINS HERE
		//*************************************************************************
		int JACKSTART;
		
		if(JACKNUM==0){
			JACKSTART=-1;
		}else{
			JACKSTART=0;
		}

		for(int JCNT=JACKSTART;JCNT<JACKNUM;JCNT++){
	    
		    //*********************************************************************
		    //	Determine the measurements to be used in the current Jacknife run
		    //*********************************************************************
			
		    if(JCNT==-1){
			    CONFL=NUMCONF;CONFU=NUMCONF;
		    }else{
			    CONFL=JCNT*NUMCONF/JACKNUM;CONFU=(JCNT+1)*NUMCONF/JACKNUM;
		    }
		    
		    
		    if(MOCKFUNC<0) {CONFL=MOCKFUNC;CONFU=abs(MOCKFUNC);
		    
		    fprintf(stderr, "\n**********************************************************\n");
		    fprintf(stderr, "***   SLICE DX:%d JACKNIFE ROUND %d OF %d START!     ***\n",DX,JCNT+1,JACKNUM+1);
		    fprintf(stderr, "***   CONFIGS:%4d USING EVERY %d CONFIGS             ***\n",NUMCONF/CONFU,CONFU);
		    fprintf(stderr, "**********************************************************\n");
		    
		    }else{

		    fprintf(stderr, "\n**********************************************************\n");
		    fprintf(stderr, "***   SLICE DX:%d JACKNIFE ROUND %d OF %d START!     ***\n",DX,JCNT+1,JACKNUM+1);
		    fprintf(stderr, "***   CONFIGS:%4d EXCL.BLOCK [%4d,%4d]              ***\n",NUMCONF-(CONFU-CONFL),CONFL,CONFU);
		    fprintf(stderr, "**********************************************************\n");
		    }
	    
		    //*********************************************************************
		    //	Allocate all vectors that are needed for the current run
		    //*********************************************************************
		    
		    A=gslgmp_vector_calloc(NOMEGA);
		    Adw=gslgmp_vector_calloc(NOMEGA);
		    DA=gslgmp_vector_calloc(TWIDTH);
		    DAWCOR=gslgmp_vector_calloc(TWIDTH);

		    Corr=gslgmp_matrix_calloc(TWIDTH,TWIDTH);
		    CEvec=gslgmp_matrix_calloc(TWIDTH,TWIDTH);
		    CEval=gslgmp_vector_calloc(TWIDTH);
		    DCDXData=gslgmp_vector_calloc(TWIDTH);
		    
		    //*************************************************************************
		    //	DATA ACQUISITION PART:   
		    //
		    //	If we use real data, call the DATA_ACQUISITION ROUTINE, which reads
		    //	the data from the harddisk and also generates the Correlation Matrix.
		    //
		    //	If we want to generate mock data we call the DATA_ACQUISITION_MOCK 
		    //	routine
		    //
		    //*************************************************************************
			
		    #if (WITHGUI==1) 
			wxMutexGuiEnter();
			*TInfo->threadtxt << wxT("Jacknife Round: [") << JCNT << wxT("] Data Acquisition for DX: [") << DX << wxT("] \n");
		        wxMutexGuiLeave();
		    #endif
			
		    DATACQ_PARM DataAcqParameter;
		    
			DataAcqParameter.NUMCONF=NUMCONF;
			DataAcqParameter.NT=NT;
			DataAcqParameter.TWIDTH=TWIDTH;
			DataAcqParameter.FILEFORMAT=FILEFORMAT;
			DataAcqParameter.lat_fname=lat_fname;
			DataAcqParameter.config_path=config_path;
			DataAcqParameter.FILENAME=FILENAME;
			DataAcqParameter.FOFFSET=FOFFSET;
			DataAcqParameter.XMAX=XMAX;
			DataAcqParameter.TREADMAX=TREADMAX;
			DataAcqParameter.NTMIN=NTMIN;
			DataAcqParameter.NTMAX=NTMAX;
			DataAcqParameter.tau=tau;
			DataAcqParameter.NOMEGA=NOMEGA;
			DataAcqParameter.MOCKFUNC=MOCKFUNC;
			DataAcqParameter.MOCKNOISE=MOCKNOISE;
			DataAcqParameter.KERNELTYPE=KERNELTYPE;
				
			DataAcqParameter.Corr=Corr;
			DataAcqParameter.CEvec=CEvec;
			DataAcqParameter.CEval=CEval;
			DataAcqParameter.DXData=DXData;

			DataAcqParameter.DX=DX;
			DataAcqParameter.CONFL=CONFL;
			DataAcqParameter.CONFU=CONFU;

			DataAcqParameter.WMIN=WMIN;
			DataAcqParameter.WMAX=WMAX;
			
			DataAcqParameter.mprior=mprior;
			DataAcqParameter.mpriordw=mpriordw;
			DataAcqParameter.mpriorback=mpriorback;
			
			DataAcqParameter.RESULTNAME=RESULTNAME;
			DataAcqParameter.BETA=BETA;
			DataAcqParameter.fname=fname;
			
		    if(MOCK==0){
			    
		       #if (WITHGUI==1) 
			    DATA_ACQUISITION(DataAcqParameter,TInfo);
		       #else
			    DATA_ACQUISITION(DataAcqParameter);
		       #endif
		    
		    } else{
			    DATA_ACQUISITION_MOCK(DataAcqParameter);
		    }
			
		    //*************************************************************************
		    //	IF WE DO NOT USE THE TRICK OF TRANSFORMING THE DATA INTO THE 
		    //	DECORRELATED BASIS, WE NEED TO COMPUTE THE INVERSE OF THE CORRELATION
		    //	MATRIX EXPLICITLY
		    //*************************************************************************

		    #if (DECORR==0)

			    gmpQRInvert(Corr,CInv);
			    INV_test(Corr,CInv);
			    SYM_test(CInv);
    
		    #endif

		    //*************************************************************************
		    //	PRINT THE RAW, CORRELATED DATA ONTO THE SCREEN
		    //*************************************************************************
    
		    fprintf(stderr, "\n**********************************************************\n");
		    fprintf(stderr, "***   SLICE DX:%d CORRELATED DATA SET                  ***\n",DX);
		    fprintf(stderr, "**********************************************************\n");
		    gmpPVec(DXData);
			
		    //*************************************************************************
		    //	DECORRELATE THE KERNEL AND DATA
		    //*************************************************************************
    
		    gslgmp_matrix_free(DCKernel);
		    DCKernel=gslgmp_matrix_calloc(Kernel->size1,Kernel->size2);
    
		    gmp_mult2m(CEvec,1,Kernel,0,DCKernel);
    
		    gslgmp_vector_free(DCDXData);
		    DCDXData=gslgmp_vector_calloc(DXData->size);
    
		    gmp_multmv(CEvec,1,DXData,DCDXData);
    
		    
		    //*************************************************************************
		    //	PRINT THE DECORRELATED DATA ONTO THE SCREEN
		    //*************************************************************************
		    
  		    fprintf(stderr, "\n**********************************************************\n");
		    fprintf(stderr, "***   SLICE DX:%d DECORRELATED DATASET                 ***\n",DX);
		    fprintf(stderr, "**********************************************************\n");
		    gmpPVec(DCDXData);
		    
		    
		    //*************************************************************************
		    //	OUTPUT THE ORIGINAL LATTICE DATA AND THE DECORRELATED DATA
		    //*************************************************************************
		    
		    sprintf(fname, "%s%s", RESULTNAME, "MEM_DCDXDATA.dat");
		    avgout=fopen(fname,"a");

		    for(int dt=NTMIN;dt<=NTMAX;dt++){
			    fprintf(avgout,"%d %.15e %.15e %.15e %.15e %.15e\n",DX+1,gslgmp_vector_get_d(tau,dt-NTMIN),gslgmp_vector_get_d(DCDXData,dt-NTMIN),sqrt(gslgmp_vector_get_d(CEval,dt-NTMIN)),sqrt(gslgmp_matrix_get_d(Corr,dt-NTMIN,dt-NTMIN)),gslgmp_vector_get_d(DXData,dt-NTMIN));
		    }
		    fclose(avgout);
			
		    fprintf(stderr, "\n**********************************************************\n");
		    fprintf(stderr, "***   SLICE DX:%d DATA ACQUISITION COMPLETE            ***\n",DX);
		    fprintf(stderr, "**********************************************************\n");
		    
		    //*************************************************************************
		    //	END OF DATA ACQUISITION PART                               
		    //*************************************************************************
		    
		    //*************************************************************************
		    //	CALCULATE THE SVD OF THE KERNEL:
		    //
		    //	THIS CAN BE USED TWICE IN THE MEM ALGORITHM: FIRST IT CAN BE USED
		    //	THE ACCELERATE THE CALCULATION OF THE EIGENVALUES OF THE LAMBDA MATRIX
		    //	AND SECONDLY IT CAN ALSO BE USED IN RESTRICTING THE SEARCH SPACE
		    //	OF THE MINIMIZER.
		    //
		    //*************************************************************************
    
		    #if (WITHGUI==1) 
		        wxMutexGuiEnter();
			*TInfo->threadtxt << wxT("Jacknife Round: [") << JCNT << wxT("] SVD Decomposition for DX: [") << DX << wxT("] \n") ;
		        wxMutexGuiLeave();
		    #endif
		    
		    gmpSVD_Transpose(DCKernel, U, V, Xi, &NSING);			
		    SVDTrans_test(DCKernel,U,V,Xi);

		    //*************************************************************************
		    //	CONSTRUCT THE MATRICES U,V, XI THAT ENCODE THE SINGULAR SUBSPACE
		    //
		    //	IN CASE THAT WE DECIDE TO USE AN EXTENDED SEARCH SPACE WE HAVE TO 
		    //	ALSO GENERATE THE CORRESPONDING EXTENDED VERSIONS OF U,V,XI
		    //
		    //*************************************************************************

		    double tw=TWIDTH;
		    int CHNS= TWIDTH;
    
		    fprintf(stderr, "\n**********************************************************\n");
		    fprintf(stderr, "***   SINGULAR SUBSPACE IS %3d ELEMENTS LARGE          ***\n",NSING);
		    fprintf(stderr, "***   AND WE ARE USING %3d ELEMENTS OUT OF THESE       ***\n",CHNS);
		    fprintf(stderr, "**********************************************************\n");

			
			//*************************************************************************
			//	HERE STARTS THE USUAL SINGULAR SUBSPACE PART
			//*************************************************************************


			NSING=CHNS;
			
			Us=gslgmp_matrix_calloc(NOMEGA,NSING);
			Vs=gslgmp_matrix_calloc(TWIDTH,NSING);
			XisM=gslgmp_matrix_calloc(NSING,NSING);
			Xis=gslgmp_vector_calloc(NSING);
			
			for(int w=0;w<NSING;w++) mpfr_set(Xis->data[w],Xi->data[w],GMP_RNDN);
			
			gmpPVec(Xis);
			
			for(int i=0;i<NOMEGA;i++)
				for(int j=0;j<NSING;j++){
					gslgmp_matrix_get(U,i,j,ptmp1);
					gslgmp_matrix_set(Us,i,j,ptmp1);
			}

			for(int i=0;i<TWIDTH;i++)
				for(int j=0;j<NSING;j++){
					gslgmp_matrix_get(V,i,j,ptmp1);
					gslgmp_matrix_set(Vs,i,j,ptmp1);
			}

			gmp_VMakeM(Xis,XisM);

			//*************************************************************************
			//	HERE STARTS THE EXTENDED SEARCH SPACE PART (ALL NAMES CARRY Ext)
			//*************************************************************************
		    
			if(ALGORITHM==9||ALGORITHM==10){

				UsExt=gslgmp_matrix_calloc(NOMEGA,TWIDTH+SVDEXTEND);
			
				if(SEARCHBASIS>0){
					
					gslgmp_vector* tauext, *XiExt;
					gslgmp_matrix* KernelExt,*VExt;
					
					XiExt=gslgmp_vector_calloc(TWIDTH+SVDEXTEND);
					tauext=gslgmp_vector_calloc(TWIDTH+SVDEXTEND);				
					KernelExt=gslgmp_matrix_calloc(TWIDTH+SVDEXTEND,NOMEGA);
					VExt=gslgmp_matrix_calloc(TWIDTH+SVDEXTEND,TWIDTH+SVDEXTEND);

				
					if(MYRANK==0){
					fprintf(stderr, "\n**********************************************************\n");
					fprintf(stderr, "***   EXTENDING THE SINGULAR SUBSPACE SEARCH BY USING A***\n");
					fprintf(stderr, "***   KERNEL THAT COMES FROM %d   MORE TAU STEPS       ***\n",SVDEXTEND);
					fprintf(stderr, "**********************************************************\n");
					}
					
					for(int dt=NTMIN;dt<=NTMAX+SVDEXTEND;dt++){
				
						gslgmp_vector_set_d(tauext,dt-NTMIN,BETA*((double)dt)/((double)(NT+SVDEXTEND)));
					}


					if(KERNELTYPE==1){
					
					//*************************************************************
					//		Zero temperature Kernel
					//		K(w_i,t_j)=Exp(-w_i*t_j)
					//*************************************************************

					mpfr_t ktmp;
					mpfr_init2(ktmp,PRECBITS);

					for(int dt=0;dt<TWIDTH+SVDEXTEND;dt++){
					for(int dw=0;dw<NOMEGA;dw++){

						gslgmp_vector_get(tauext,dt,ptmp1);
						gslgmp_vector_get(omega,dw,ptmp2);
						mpfr_mul(ptmp3,ptmp1,ptmp2,GMP_RNDN);
						mpfr_neg(ptmp1,ptmp3,GMP_RNDN);
						mpfr_exp (ktmp, ptmp1, GMP_RNDN);

						gslgmp_matrix_set(KernelExt,dt,dw,ktmp);


					}}
					} else if(KERNELTYPE==2){
					
					//*************************************************************
					//		Naive finite temperature Kernel(Half BETA)
					//		K(w_i,t_j)= Cosh[w(t-0.5\beta)]/Sinh[w*0.5\beta]
					//*************************************************************
					
					mpfr_t ktmp;
					mpfr_init2(ktmp,PRECBITS);

					for(int dt=0;dt<TWIDTH+SVDEXTEND;dt++){
					for(int dw=0;dw<NOMEGA;dw++){

						gslgmp_vector_get(tauext,dt,ptmp1);
						gslgmp_vector_get(omega,dw,ptmp2);
						
						mpfr_sub_d(ptmp3, ptmp1,0.5*BETA,GMP_RNDN);
						mpfr_mul(ptmp3,ptmp3,ptmp2,GMP_RNDN);
						mpfr_cosh(ptmp1,ptmp3,GMP_RNDN);
						
						mpfr_mul_d(ptmp3,ptmp2,0.5*BETA,GMP_RNDN);
						mpfr_sinh(ptmp2,ptmp3,GMP_RNDN);

						mpfr_div(ptmp3,ptmp1,ptmp2,GMP_RNDN);
						
						gslgmp_matrix_set(KernelExt,dt,dw,ptmp3);
						

					}}
					
					} else if(KERNELTYPE==3){
					
					//*************************************************************
					//		Regularized finite temperature Kernel
					//		K(w_i,t_j)= Tanh[omega\beta/2]*Cosh[w(t-0.5\beta)]/Sinh[w*0.5\beta]
					//                    = Cosh[w(t-0.5\beta)]/Cosh[\omega\beta/2]
					//*************************************************************
					
					mpfr_t ktmp;
					mpfr_init2(ktmp,PRECBITS);

					for(int dt=0;dt<TWIDTH+SVDEXTEND;dt++){
					for(int dw=0;dw<NOMEGA;dw++){

						gslgmp_vector_get(tauext,dt,ptmp1);
						gslgmp_vector_get(omega,dw,ptmp2);
						
						mpfr_sub_d(ptmp3, ptmp1,0.5*BETA,GMP_RNDN);
						mpfr_mul(ptmp3,ptmp3,ptmp2,GMP_RNDN);
						mpfr_cosh(ptmp1,ptmp3,GMP_RNDN);
						
						mpfr_mul_d(ptmp3,ptmp2,0.5*BETA,GMP_RNDN);
						mpfr_cosh(ptmp2,ptmp3,GMP_RNDN);

						mpfr_div(ptmp3,ptmp1,ptmp2,GMP_RNDN);
						
						gslgmp_matrix_set(KernelExt,dt,dw,ptmp3);
						

					}}
		
					}else{
					
						
						fprintf(stderr, "\n**********************************************************\n");
						fprintf(stderr, "***   PLEASE CHOOSE VALID KERNEL 1 OR 2       exiting..***\n");
						fprintf(stderr, "**********************************************************\n");
						exit(-1);
						
					}
					
					

					//*************************************************************************
					//	SHOW THE EXTENDED TAU VALUES
					//*************************************************************************

					gmpPVec(tauext);
					
					
					gmpSVD_Transpose(KernelExt, UsExt, VExt, XiExt, &NSINGEXT);			
					SVDTrans_test(KernelExt,UsExt,VExt,XiExt);
	
					//*************************************************************************
					//	ONE SHOULD REPLACE THE FIRST TWIDTH COLUMS OF UsExt
					//	BY THE ORIGINAL SVD BASIS VECTORS FROM THE DECORRELATED KERNEL
					//*************************************************************************
					
					
					for(int i=0;i< ((SVDEXTEND>0)?TWIDTH:TWIDTH+SVDEXTEND) ;i++){
						
						for(int w=0; w<NOMEGA; w++){
							gslgmp_matrix_get(Us,w,i,ptmp1);
							gslgmp_matrix_set(UsExt,w,i,ptmp1);
							
						}
					}


					gslgmp_vector_free(XiExt);
					gslgmp_vector_free(tauext);
					gslgmp_matrix_free(KernelExt);
					gslgmp_matrix_free(VExt);



				} else {
					
					if(MYRANK==0){
					fprintf(stderr, "\n**********************************************************\n");
					fprintf(stderr, "***   THIS MEM USES AS SEARCH SPACE AN                 ***\n");
					fprintf(stderr, "***   OSCIALLATING FOURIER BASIS WITH %3d STEPS        ***\n",TWIDTH+SVDEXTEND);
					fprintf(stderr, "**********************************************************\n");
					}

				    //*************************************************************
				    //		Oscillating Kernel from Fourier Basis
				    //		K(w_i,t_j)=sin(w*t*2*Pi/Nw) bzw cos(w*t*2*Pi/Nw)
				    //*************************************************************

				    mpfr_t ktmp;
				    mpfr_init2(ktmp,PRECBITS);

				    for(int dw=0;dw<NOMEGA;dw++){
					    	gslgmp_matrix_set_d(UsExt,dw,0, 1. );
				    }

				    for(int dt=2;dt<TWIDTH+SVDEXTEND+1;dt++){
				    for(int dw=0;dw<NOMEGA;dw++){

					    if(dt%2==0){
					    	gslgmp_matrix_set_d(UsExt,dw,dt-1,sin( (dt/2)*dw*(2.*PI)/( (double)NOMEGA) ) );
					    }else{
					    	gslgmp_matrix_set_d(UsExt, dw,dt-1, cos( (dt/2)*dw*(2.*PI)/( (double)NOMEGA) ));
					    }

				    }}
				}
 				
				NSINGEXT=NSING+SVDEXTEND;
			
			}

			//*************************************************************************
			//  END OF THE EXTENDED SEARCH SPACE GENERATION
			//*************************************************************************

			//*************************************************************************
			//	SINCE THE SVD IS OFTEN USED TO RESTRICT THE SEARCH SPACE OF THE 
			//	MINIMIZER THIS DEBUGGING OPTION ALLOWS TO DUMP THE CORRESPONDING
			//	SVD BASIS INTO A FILE TO ANALYZE WHETHER THE RESTRICTION IS OK OR NOT
			//*************************************************************************
			
			#if (SVDOUT>0)
			
				gslgmp_vector* bout;
				gslgmp_vector* aout;
				gslgmp_vector* Aout;
			
				char svdfname[256];
			
				sprintf(svdfname, "%sSVDBASIS_NT%d_NW%d_FULL.dat",RESULTNAME,NT,NOMEGA);
				
				FILE *SVDOUTPUTS;
				
				SVDOUTPUTS=fopen(svdfname, "w");
				
				bout=gslgmp_vector_calloc(NSING);	
				aout=gslgmp_vector_calloc(NOMEGA);	
				Aout=gslgmp_vector_calloc(NOMEGA);	
				
				for(int i=0;i<NSING;i++){

					for(int k=0;k<NSING;k++) mpfr_set_ui(bout->data[k],0,GMP_RNDN);
					
					mpfr_set_ui(bout->data[i],1,GMP_RNDN);
					
					gmp_multmv(U,0,bout,aout);
							
					for(int k=0;k<NOMEGA;k++) fprintf(SVDOUTPUTS, "%d %e %e\n", i, domega[k], mpfr_get_d(aout->data[k],GMP_RNDN));
				}
				
				
				fclose(SVDOUTPUTS);	
				
				
				if(ALGORITHM==9||ALGORITHM==10){
				
					gslgmp_vector_free(bout);
					bout=gslgmp_vector_calloc(NSINGEXT);	
					
					sprintf(svdfname, "%sEXTSVDBASIS_NT%d_NW%d_FULL.dat",RESULTNAME,NSINGEXT,NOMEGA);
					
					SVDOUTPUTS=fopen(svdfname, "w");
					
					for(int i=0;i<NSINGEXT;i++){

						for(int k=0;k<NSINGEXT;k++) mpfr_set_ui(bout->data[k],0,GMP_RNDN);
						
						mpfr_set_ui(bout->data[i],1,GMP_RNDN);
						
						gmp_multmv(UsExt,0,bout,aout);
														
						for(int k=0;k<NOMEGA;k++) fprintf(SVDOUTPUTS, "%d %e %e\n", i, domega[k], mpfr_get_d(aout->data[k],GMP_RNDN));
					}
				
				}
				
				fclose(SVDOUTPUTS);	
				
				
				gslgmp_vector_free(bout);
				gslgmp_vector_free(aout);
				gslgmp_vector_free(Aout);
				
			#endif
			#if (SVDOUT==2)
			exit(-1);
			#endif


			//*************************************************************************
			//  Allocate the files for the saving the minimizer progress
			//*************************************************************************

			FILE* ITERVAL, *PROJVAL;
			
			sprintf(fname, "%s%s", RESULTNAME, "Intervals.dat");
			ITERVAL=fopen(fname,"a");
			sprintf(fname, "%s%s", RESULTNAME, "Projvals.dat");
			PROJVAL=fopen(fname,"a");

			//*************************************************************************
			//  PREPARE FOR THE MINIMIZER ROUTINES            
			//*************************************************************************

			mpfr_set_ui(ptmp1,0,GMP_RNDN);
			
			if(ALGORITHM==2||ALGORITHM==4){
				
				blbfgs=gslgmp_vector_calloc(NOMEGA);
				b=gslgmp_vector_calloc(NOMEGA);	
				blbfgsback=gslgmp_vector_calloc(NOMEGA);
			
			}else if (ALGORITHM==0||ALGORITHM==1||ALGORITHM==3||ALGORITHM==5||ALGORITHM==6||ALGORITHM==7||ALGORITHM==8){
				
				blbfgsSVD=gslgmp_vector_calloc(NSING);
				b=gslgmp_vector_calloc(NSING);	
				blbfgsback=gslgmp_vector_calloc(NSING);
				blmback=gslgmp_vector_calloc(NSING);	
				blm=gslgmp_vector_calloc(NSING);	
				
			}else if (ALGORITHM==9||ALGORITHM==10){
				
				blbfgsSVD=gslgmp_vector_calloc(NSINGEXT);
				b=gslgmp_vector_calloc(NSINGEXT);	
				blbfgsback=gslgmp_vector_calloc(NSINGEXT);
				blmback=gslgmp_vector_calloc(NSINGEXT);	
				blm=gslgmp_vector_calloc(NSINGEXT);	

			}
			
			
			mpfr_t EPS1,EPS2,EPS3,Lnew,Lold,Lval,Lderiv,Sval,Sderiv,Qnew,Qold,Sold,Snew;
		
			mpfr_init2(EPS1,PRECBITS);
			mpfr_init2(EPS2,PRECBITS);
			mpfr_init2(EPS3,PRECBITS);
			mpfr_init2(Lnew,PRECBITS);
			mpfr_init2(Lold,PRECBITS);
			mpfr_init2(Qnew,PRECBITS);
			mpfr_init2(Qold,PRECBITS);
			mpfr_init2(Lval,PRECBITS);
			mpfr_init2(Lderiv,PRECBITS);
			mpfr_init2(Sval,PRECBITS);
			mpfr_init2(Sold,PRECBITS);
			mpfr_init2(Snew,PRECBITS);
			mpfr_init2(Sderiv,PRECBITS);
	
			mpfr_set_ui(EPS1,0,GMP_RNDN);
			mpfr_set_ui(EPS2,0,GMP_RNDN);
			mpfr_set_ui(EPS3,BFGSPREC,GMP_RNDN);
	
			int LTOLCHK=0;
			int iter=0;
	
			double alpha;
	
			double **AVEC,*FINALA,*FINALERR,**ERRS,*Lvals,*Qvals;
			
			
			AVEC=(double**)malloc(sizeof(double*)*ANUM);
			for(int i=0;i<ANUM;i++) AVEC[i]=(double*)malloc(sizeof(double)*NOMEGA);
			ERRS=(double**)malloc(sizeof(double*)*ANUM);
			for(int i=0;i<ANUM;i++) ERRS[i]=(double*)malloc(sizeof(double)*ERRNUM);
			Lvals=(double*)malloc(sizeof(double)*ANUM);
			Qvals=(double*)malloc(sizeof(double)*ANUM);
			FINALA=(double*)malloc(sizeof(double)*NOMEGA);
			FINALERR=(double*)malloc(sizeof(double)*NOMEGA);

			LOGPROB=gslgmp_vector_calloc(ANUM);
			
			
			a=gslgmp_vector_calloc(NOMEGA);

			#if (ITERPRIOR==1)

			    //*************************************************************************
			    //	ITERATIVE PRIOR:
			    //   AFTER RECONSTRUCTING THE SPECTRAL FUNCTION ONCE 
			    //   USING THE SET PRIOR WE TAKE THIS RECONSTRUCTION
			    //   AS THE NEW PRIOR, THIS IS REPEATED PRIORREP TIMES 
			    //
			    //	HERE WE READ IN THE ORIGINAL PRIOR SINCE WE ARE AT THE START OF
			    //	A NEW DX DATA SLICE
			    //*************************************************************************

			    int PRIORREP=GLOBALPRIORREP;

			    for(int dw=0;dw<NOMEGA;dw++){
	    
				    mpfr_set(mprior->data[dw],mpriorback->data[dw],GMP_RNDN);
	    
			    }

			    append_dw(mpriordw,mprior,omega,NOMEGA,HPNUM,OMEGAHPMIN,OMEGAHPMAX);

			#else
			    int PRIORREP=1;
			#endif

				
			//*************************************************************************
			//	BEGIN THE PRIOR REPETITION ROUTINE HERE
			//*************************************************************************

			for(int REP=0;REP<PRIORREP;REP++){

				fprintf(stderr, "\n**********************************************************\n");
				fprintf(stderr, "***   SLICE DX:%d PRIOR ITERATION %d of %d             ***\n",DX,REP+1,PRIORREP);
				fprintf(stderr, "**********************************************************\n");
		
				if(REP!=0){
	
					//*************************************************************************
					//	IF THIS IS NOT THE FIRST REPETITION SET THE PRIOR AS THE LAST
					//	RECONSTRUCTED SPECTRAL FUNCTION
					//*************************************************************************
	
					for(int dw=0;dw<NOMEGA;dw++){
			
						mpfr_set_d(mprior->data[dw],FINALA[dw],GMP_RNDN);
			
					}

					append_dw(mpriordw,mprior,omega,NOMEGA,HPNUM,OMEGAHPMIN,OMEGAHPMAX);
				}
			    
				//*************************************************************************
				//	OUTPUT THE PRIOR TO HD
				//*************************************************************************
				
				sprintf(fname, "%s%s", RESULTNAME, "Prior.dat");
				avgout=fopen(fname,"a");
				for(int dw=0;dw<NOMEGA;dw++){
					fprintf(avgout,"%d %d %e %.15e %.15e \n",DX+1,REP,domega[dw],mpfr_get_d(mprior->data[dw],GMP_RNDN),mpfr_get_d(mprior->data[dw],GMP_RNDN)/pow(domega[dw],PRIORPOWER));
				}
				fclose(avgout);
			

				//*************************************************************************
				//	SETUP THE INITIAL VALUES FOR THE MINIMIZER AS RANDOM 
				//	VALUES.
				//*************************************************************************
				fprintf(stderr, "\n**********************************************************\n");
				fprintf(stderr, "***   START VECTOR FOR THE MINIMIZER                   ***\n");
				fprintf(stderr, "**********************************************************\n");
				
				    if(ALGORITHM==2||ALGORITHM==4){
					    
					    for(int i=0;i<NOMEGA;i++){
						    mpfr_set_d(blbfgs->data[i],(4.*((double)(rand()%100))/100.),GMP_RNDN);
					    }
						    gmpPVec(blbfgs);
				    
				    }else if (ALGORITHM==0||ALGORITHM==1){

					    for(int i=0;i<NSING;i++){
						    mpfr_set_d(blm->data[i],(4.*((double)(rand()%100))/100.),GMP_RNDN);
					    }
						    gmpPVec(blm);

				    }else if (ALGORITHM==9||ALGORITHM==10){

					    for(int i=0;i<NSINGEXT;i++){
						    mpfr_set_d(blbfgsSVD->data[i],(4.*((double)(rand()%100))/100.),GMP_RNDN);
					    }
						    gmpPVec(blbfgsSVD);

				    }else if (ALGORITHM==3||ALGORITHM==5||ALGORITHM==6||ALGORITHM==7||ALGORITHM==8){
					    
					    for(int i=0;i<NSING;i++){
						    mpfr_set_d(blbfgsSVD->data[i],(4.*((double)(rand()%100))/100.),GMP_RNDN);
					    }
					    gmpPVec(blbfgsSVD);
				    }



				//*************************************************************************
				//	NOW WE HAVE TO START THE ALPHA LOOP TO SELF CONSISTENTLY FIND
				//	THE PROBABILITY DISTRIBUTION OVER WHICH WE FINALLY INTEGRATE
				//	ALL THE MEM RECONSTRUCTIONS FOR A SPECIFIC VALUE OF ALPHA
				//*************************************************************************
			    
				acnt=ANUM;
				alpha=ASTART;
		
				while (acnt>=0){				  

					//*************************************************************************
					//	Need to approach the small alpha values softly from 
					//	above otherwise minimizer might not converge
					//	THE ALPHAS ARE TRAVERSED LINEARLY
					//*************************************************************************
		
					if(acnt!=ANUM){

						alpha=AMIN+(acnt+1.)*(AMAX-AMIN)/ANUM;

					}

					#if (WITHGUI==1) 
					    wxMutexGuiEnter();
					    *TInfo->threadtxt << wxT("Jacknife Round: [") << JCNT << wxT("] running the Minimizer for DX: [") << DX << wxT("] with alpha: [") << alpha << wxT("] \n");
					    wxMutexGuiLeave();
					#endif

					mpfr_set_d(Lnew,10.,GMP_RNDN);
					mpfr_set_d(Lold,1.,GMP_RNDN);

					mpfr_set_d(Snew,10.,GMP_RNDN);
					mpfr_set_d(Sold,1.,GMP_RNDN);

					mpfr_set_d(Qnew,1e10,GMP_RNDN);
					mpfr_set_d(Qold,1.,GMP_RNDN);
		
					LTOLCHK=0;
					iter=0;
		
					fprintf(stderr, "\n**********************************************************\n");
					fprintf(stderr, "***   SLICE DX:%d WITH ALPHA: %e  %d/%d              ***\n",DX,alpha,ANUM-acnt,ANUM);
					fprintf(stderr, "**********************************************************\n");

					//*************************************************************************
					//	PERFORM THE MINIMIZATION FOR THIS ALPHA VALUE
					//*************************************************************************

					//*************************************************************************
					//	SET THE PRECISION FOR THE MINIMIZER TO THE CURRENT ONE: PRECBITS
					//*************************************************************************

					if (ALGORITHM==2){

						#if (REDOMEM==1)
						
							#include "./AUXILIARY_ROUTINES/REDO_ERRORS.c"
						
						#else 
						
							#include "./MINIMIZER_ROUTINES/LBFGS_FULLA.c"
							
						#endif

				        }

				    	if (ALGORITHM==3){

						#include "./MINIMIZER_ROUTINES/LBFGS_SVD.c"
		
				   	}

				    	if (ALGORITHM==6){

						#include "./MINIMIZER_ROUTINES/LBFGS_LM_SVD.c"

					}
					
				    	if (ALGORITHM==10){

						#include "./MINIMIZER_ROUTINES/LBFGS_LM_SVD_LESSSTRICT.c"

					}


					//*********************************************************
					//	END OF MINIMIZER SECTION
					//*********************************************************
		
					if(acnt<ANUM){

						//*********************************************************
						//	Calculate the final A vector for this alpha
						//*********************************************************

						if (ALGORITHM==2||ALGORITHM==4){
							
							#if (REDOMEM==1)
							
								for(int dw=0;dw<NOMEGA;dw++){
									mpfr_set(A->data[dw],blbfgs->data[dw],GMP_RNDN);
								}
								
							#else
							
								for(int dw=0;dw<NOMEGA;dw++){
									mpfr_set(ptmp2,blbfgs->data[dw],GMP_RNDN);
									mpfr_exp(A->data[dw],ptmp2,GMP_RNDN);
								}
							
							#endif
								
						}

						if (ALGORITHM==3||ALGORITHM==5||ALGORITHM==6||ALGORITHM==7||ALGORITHM==8){

							gmp_multmv(Us,0,blbfgsSVD,a);

							for(int dw=0;dw<NOMEGA;dw++){
								mpfr_set(ptmp2,a->data[dw],GMP_RNDN);
								mpfr_exp(A->data[dw],ptmp2,GMP_RNDN);
								mpfr_mul(A->data[dw],A->data[dw],mprior->data[dw],GMP_RNDN);
							}
						}

						if (ALGORITHM==9||ALGORITHM==10){

							gmp_multmv(UsExt,0,blbfgsSVD,a);

							for(int dw=0;dw<NOMEGA;dw++){
								mpfr_set(ptmp2,a->data[dw],GMP_RNDN);
								mpfr_exp(A->data[dw],ptmp2,GMP_RNDN);
								mpfr_mul(A->data[dw],A->data[dw],mprior->data[dw],GMP_RNDN);
							}
						}

						if (ALGORITHM==0||ALGORITHM==1){

							gmp_multmv(Us,0,blm,a);

							for(int dw=0;dw<NOMEGA;dw++){
								mpfr_set(ptmp2,a->data[dw],GMP_RNDN);
								mpfr_exp(A->data[dw],ptmp2,GMP_RNDN);
								mpfr_mul(A->data[dw],A->data[dw],mprior->data[dw],GMP_RNDN);
							}
						}
						
						//*********************************************************
						//	ADD THE NECESSARY DW FACTORS FOR THE INTEGRAL
						//*********************************************************

						append_dw(Adw,A,omega,NOMEGA,HPNUM,OMEGAHPMIN,OMEGAHPMAX);
					
						//*********************************************************
						//	SAVE THE RECONSTRUCTION FOR THIS ALPHA
						//*********************************************************

						for(int dw=0;dw<NOMEGA;dw++){
							AVEC[acnt][dw]=mpfr_get_d(A->data[dw],GMP_RNDN);
						}

						//*********************************************************
						//	Generate data from the current A
						//*********************************************************
					
						gmp_multmv(DCKernel,0,Adw,DA);

						//*********************************************************
						//	Generate correlated data from the current A
						//*********************************************************
					
						gmp_multmv(Kernel,0,Adw,DAWCOR);
			
						#if (WITHGUI==1) 
						    wxMutexGuiEnter();
						    *TInfo->threadtxt << wxT("Jacknife Round: [") << JCNT << wxT("] Errors and Probability for DX: [") << DX << wxT("] with alpha: [") << alpha << wxT("] \n");
						    wxMutexGuiLeave();
						#endif
		
						fprintf(stderr, "\n**********************************************************\n");
						fprintf(stderr, "****       ERROR ESTIMATION FOR THIS ALPHA  %4d  ********\n",ANUM-acnt);
						fprintf(stderr, "**********************************************************\n");
		
						//*********************************************************
						//	Calculate the Error estimates over ERRNUM intervals
						//	along the omega direction using local Q's
						//
						//	i.e.: d^2Q/dAdA=+alpha*d2SdAdA-d2LdAdA
						//*********************************************************
						
						//*********************************************************
						//	Calculate the local error submatrices
						//*********************************************************

						if (FULLERR==0){
							#include "./AUXILIARY_ROUTINES/ERROR_TRICK.c"
						}else{
							#include "./AUXILIARY_ROUTINES/ERROR_FULL.c"
						}
					
						//*********************************************************
						//	Calculate the probability for this a and alpha
						//*********************************************************
			
						fprintf(stderr, "\n**********************************************************\n");
						fprintf(stderr, "**** PROBABILITY ESTIMATION FOR THIS ALPHA  %4d  ********\n",ANUM-acnt);
						fprintf(stderr, "**********************************************************\n");

						//*************************************************************************
						//	SET THE PRECISION FOR THE LAMBDA EIGENVALUE DETERMINATION 
						//	TO THE CHODEN HIGHER PRECISION
						//*************************************************************************
						
						#if (LTRICK==0)
							#include "./AUXILIARY_ROUTINES/LAMBDA_FULL.c"
						#endif
						
						#if (LTRICK==1)
							#include "./AUXILIARY_ROUTINES/LAMBDA_TRICK.c"
						#endif
												
						//*********************************************************
						//	END OF EIGENVALUE SECTION FOR LAMBDA
						//*********************************************************

						Lvals[acnt]=mpfr_get_d(Lnew,GMP_RNDN);
						Qvals[acnt]=mpfr_get_d(Qnew,GMP_RNDN);
						
						//*********************************************************
						//	OUTPUT THE CURRENT RECONSTRUCTION FOR THIS ALPHA
						//*********************************************************

						sprintf(fname, "%s%s", RESULTNAME, "MEM_spec_rec_alpha.dat");
						rho_output=fopen(fname,"a");

						if (JACKNUM==0){
							if(acnt<ANUM){
								for(int w=0; w<NOMEGA; w++){
								fprintf(rho_output,"%d %e %d %e %e %e %e %e %e %d %d %e %e %e\n", DX,AMIN+(acnt+1.)*(AMAX-AMIN)/ANUM,w,mpfr_get_d(omega->data[w],GMP_RNDN), GNUPR(gslgmp_vector_get_d(A,w))/pow(domega[w],PRIORPOWER),ERRS[acnt][w/(NOMEGA/ERRNUM)]/pow(domega[w],PRIORPOWER),GNUPR(gslgmp_vector_get_d(A,w)),ERRS[acnt][w/(NOMEGA/ERRNUM)],mpfr_get_d(LOGPROB->data[acnt],GMP_RNDN),JCNT,REP,GNUPR(gslgmp_vector_get_d(Adw,w)),Qvals[acnt],Lvals[acnt] );
							}}
						}

						fflush(rho_output);

						sprintf(fname, "%s%s", RESULTNAME, "MEM_datacomp.dat");
						data_output=fopen(fname,"a");
						if(acnt<ANUM){
						for(int t=0; t<TWIDTH; t++){
						fprintf(data_output,"%d %e %d %e %e %e %e %d %d\n", DX,AMIN+(acnt+1.)*(AMAX-AMIN)/ANUM,t,mpfr_get_d(tau->data[t],GMP_RNDN), gslgmp_vector_get_d(DA,t),gslgmp_vector_get_d(DAWCOR,t),mpfr_get_d(LOGPROB->data[acnt],GMP_RNDN),JCNT,REP);
						}}
			
						fclose(rho_output);
						fclose(data_output);
			
			
						fprintf(stderr, "\n**********************************************************\n");
						fprintf(stderr, "***   RESULT ALPHA:%e Q:%e L:%e S:%e Log(P):%e          ***\n",alpha,-mpfr_get_d(Lval,GMP_RNDN)+alpha*mpfr_get_d(Sval,GMP_RNDN),mpfr_get_d(Lval,GMP_RNDN),mpfr_get_d(Sval,GMP_RNDN),mpfr_get_d(LOGPROB->data[acnt],GMP_RNDN));
						fprintf(stderr, "**********************************************************\n");
						
					}
					
					acnt--;
		
					#if (WITHGUI)
					   wxMutexGuiEnter();
					   TInfo->threadgauge->SetValue( 100.*( (DX-STARTDX)/(XMAX-STARTDX+1.) ) + 100.*(1./(XMAX-STARTDX+1.))*( (ANUM-acnt)/ANUM ) );
					   wxMutexGuiLeave();
					#endif
					
				} 
				
				//*********************************************************
				//	END OF THE ALPHA LOOP
				//*********************************************************

				fclose(ITERVAL);
				fclose(PROJVAL);			    
			    
				//*********************************************************
				//	Sum the probabilities for alpha
				//	and normalize to one
				//*********************************************************
		
				gslgmp_vector* NPROB,*UPROB;
		
				NPROB=gslgmp_vector_calloc(ANUM);
				UPROB=gslgmp_vector_calloc(ANUM);
		
				mpfr_t NORMA;
				mpfr_init2(NORMA,PRECBITS);
				mpfr_set_ui(NORMA,0,GMP_RNDN);
		
				for(int i=0;i<ANUM;i++) { mpfr_exp(UPROB->data[i],LOGPROB->data[i],GMP_RNDN);}
		
				for(int i=0;i<ANUM;i++) {mpfr_add(NORMA,NORMA,UPROB->data[i],GMP_RNDN);}
		
				for(int i=0;i<ANUM;i++) {mpfr_div(NPROB->data[i],UPROB->data[i],NORMA,GMP_RNDN);}

		
				//*********************************************************
				//	OUTPUT ALPHA PROBABILITIES
				//*********************************************************
				
					sprintf(fname, "%s%s", RESULTNAME, "Probabilities.dat");
					rho_output=fopen(fname,"a");
			
					for(int i=0; i<ANUM; i++){
						fprintf(rho_output,"%d %d %d %e %.15e %.15e %d %d\n", DX,-1,i,AMIN+i*(AMAX-AMIN)/ANUM,mpfr_get_d(UPROB->data[i],GMP_RNDN),mpfr_get_d(NPROB->data[i],GMP_RNDN),JCNT,REP);
					}
					fclose(rho_output);
				
				//*********************************************************
				//	FIND THE MAXIMUM PROBABILITY FOR ALPHA
				//*********************************************************
				
				int amax=0;
				double maxp=0.;
				
				for(int i=0; i<ANUM; i++){
				
					if(mpfr_get_d(NPROB->data[i],GMP_RNDN)>maxp) { maxp = mpfr_get_d(NPROB->data[i],GMP_RNDN); amax =i;}
					
				}

					
				//*********************************************************
				//	Calculate the final A vector averaged over alpha
				//*********************************************************
		
				for(int dw=0;dw<NOMEGA;dw++){
					FINALA[dw]=0.;
					for(int i=0;i<ANUM;i++) {
						FINALA[dw]+=mpfr_get_d(NPROB->data[i],GMP_RNDN)*AVEC[i][dw];
					}
				}


				//*********************************************************
				//	Calculate the final ERROR vector averaged over alpha
				//*********************************************************
		
				for(int dw=0;dw<NOMEGA;dw++){
					FINALERR[dw]=0.;
					for(int i=0;i<ANUM;i++) {
						FINALERR[dw]+=mpfr_get_d(NPROB->data[i],GMP_RNDN)*ERRS[i][dw/(NOMEGA/ERRNUM)];
					}
				}
		
				gslgmp_vector_free(NPROB);
				gslgmp_vector_free(UPROB);
				mpfr_clear(NORMA);
				
				//*********************************************************
				//	Generate data from the final A
				//*********************************************************
		
				for(int dw=0;dw<NOMEGA;dw++){
					mpfr_set_d(A->data[dw],FINALA[dw],GMP_RNDN);
				}

				//*********************************************************
				//	ADD THE NECESSARY DW FACTORS FOR THE INTEGRAL
				//*********************************************************

				append_dw(Adw,A,omega,NOMEGA,HPNUM,OMEGAHPMIN,OMEGAHPMAX);

				gmp_multmv(DCKernel,0,Adw,DA);
				gmp_multmv(Kernel,0,Adw,DAWCOR);

		
				//*********************************************************
				//	OUTPUT THE FINAL RECONSTRUCTION AND ITS DATA 
				//*********************************************************
				
			
				sprintf(fname, "%s%s", RESULTNAME, "spec_rec.dat");
				rho_output=fopen(fname,"a");
		
				for(int w=0; w<NOMEGA; w++){
				fprintf(rho_output,"%d %d %e %e %e %e %e %d %d %e %e %e %e %e %e %d %d\n", DX,w,mpfr_get_d(omega->data[w],GMP_RNDN), GNUPR(FINALA[w])/pow(domega[w],PRIORPOWER),FINALERR[w]/pow(domega[w],PRIORPOWER),GNUPR(FINALA[w]),FINALERR[w],JCNT,REP,GNUPR(AVEC[amax][w]), GNUPR(ERRS[amax][w/(NOMEGA/ERRNUM)]), GNUPR(mpfr_get_d(Adw->data[w],GMP_RNDN)), FINALERR[w], Qvals[amax],Lvals[amax],JCNT, REP);
				}
				fclose(rho_output);
		
		
				sprintf(fname, "%s%s", RESULTNAME, "spec_recdat.dat");
				rho_output=fopen(fname,"a");
		
				for(int t=0; t<TWIDTH; t++){
				fprintf(rho_output,"%d %d %e %e %e %d %d\n", DX,t,mpfr_get_d(tau->data[t],GMP_RNDN),GNUPR(mpfr_get_d(DA->data[t],GMP_RNDN)),GNUPR(mpfr_get_d(DAWCOR->data[t],GMP_RNDN)),JCNT, REP);
				}
				fclose(rho_output);
				
			}
			
			//*********************************************************
			//	END OF PRIOR ITERATION LOOP
			//*********************************************************

			//*********************************************************
			//	Save the current final A vector averaged over alpha
			//	for further JACKNIFE ANALYSIS
			//*********************************************************
			if(JACKNUM!=0){
				
				for(int dw=0;dw<NOMEGA;dw++){
						FINALAJACK[JCNT][dw]=FINALA[dw];
						FINALERRJACK[JCNT][dw]=FINALERR[dw];
				}
				
			}

	
			mpfr_clear(EPS1);
			mpfr_clear(EPS2);
			mpfr_clear(EPS3);
			mpfr_clear(Lnew);
			mpfr_clear(Lold);
			mpfr_clear(Lval);
			mpfr_clear(Lderiv);
			mpfr_clear(Qnew);
			mpfr_clear(Qold);
			mpfr_clear(Sval);
			mpfr_clear(Sderiv);
	
			for(int i=0;i<ANUM;i++) free(AVEC[i]);
			free(AVEC);
			for(int i=0;i<ANUM;i++) free(ERRS[i]);
			free(ERRS);
			free(Qvals);
			free(Lvals);
			free(FINALA);
			free(FINALERR);
			gslgmp_vector_free(LOGPROB);
			
			if(ALGORITHM==1) gslgmp_vector_free(blbfgs);
			if(ALGORITHM==0) gslgmp_vector_free(blm);

			
			if(ALGORITHM==2||ALGORITHM==4){
				
				gslgmp_vector_free(blbfgs);
				gslgmp_vector_free(b);
				gslgmp_vector_free(blbfgsback);
			
			}else if (ALGORITHM==0||ALGORITHM==1||ALGORITHM==3||ALGORITHM==5||ALGORITHM==6||ALGORITHM==7||ALGORITHM==8){
				
				gslgmp_vector_free(blbfgsSVD);
				gslgmp_vector_free(b);
				gslgmp_vector_free(blbfgsback);
				gslgmp_vector_free(blmback);
				gslgmp_vector_free(blm);
				
			}else if (ALGORITHM==9||ALGORITHM==10){
				
				gslgmp_vector_free(blbfgsSVD);
				gslgmp_vector_free(b);
				gslgmp_vector_free(blbfgsback);
				gslgmp_vector_free(blmback);
				gslgmp_vector_free(blm);
				
			}
			
			gslgmp_vector_free(a);
			gslgmp_vector_free(A);
			gslgmp_vector_free(Adw);
			gslgmp_vector_free(DA);
			gslgmp_vector_free(DAWCOR);
			gslgmp_vector_free(Xis);
	

			gslgmp_matrix_free(Us);
			gslgmp_matrix_free(Vs);
			gslgmp_matrix_free(XisM);
			
			fprintf(stderr, "\n**********************************************************\n");
			fprintf(stderr, "***   SLICE DX:%d JACKNIFE ROUND %d OF %d DONE       ***\n",DX,JCNT+1,JACKNUM+1);
			fprintf(stderr, "**********************************************************\n");

			gslgmp_matrix_free(Corr);
			gslgmp_matrix_free(CEvec);
			gslgmp_vector_free(CEval);
			gslgmp_vector_free(DCDXData);
		}
		
		//*********************************************************
		//	END OF JACKNIFE LOOP:
		//	NOW WE HAVE TO DO THE JACKNIFE ANALYSIS BASED
		//	ON THE MEM RECONSTRUCTIONS OBTAINED FOR EACH OF
		//	MEASUREMENTS WITH ONE BLOCK REMOVED.
		//*********************************************************
		if(JACKNUM!=0){
			
			double* FINALJACKA;
			FINALJACKA=(double*)malloc(sizeof(double)*NOMEGA);
			double* FINALJACKVAR;
			FINALJACKVAR=(double*)malloc(sizeof(double)*NOMEGA);
			double* FINALJACKAERR;
			FINALJACKAERR=(double*)malloc(sizeof(double)*NOMEGA);
		
			double JNORM=1./(JACKNUM);
			double JVNORM=((double)JACKNUM-1.)/((double)(JACKNUM));

			for(int dw=0;dw<NOMEGA;dw++){

				FINALJACKA[dw]=0.;
				FINALJACKVAR[dw]=0.;
				FINALJACKAERR[dw]=0.;

				for(int i=0;i<JACKNUM;i++){
				FINALJACKA[dw]+=FINALAJACK[i][dw]*JNORM;
				}

				for(int i=0;i<JACKNUM;i++){
				FINALJACKVAR[dw]+=pow(FINALJACKA[dw]-FINALAJACK[i][dw],2)*JVNORM;
				}

				for(int i=0;i<JACKNUM;i++){
				FINALJACKAERR[dw]+=FINALERRJACK[i][dw]*JNORM;
				}
			}
	
	
			//*********************************************************
			//	Generate data from the final A
			//*********************************************************

			A=gslgmp_vector_calloc(NOMEGA);
			Adw=gslgmp_vector_calloc(NOMEGA);
			DA=gslgmp_vector_calloc(TWIDTH);
			DAWCOR=gslgmp_vector_calloc(TWIDTH);

			for(int dw=0;dw<NOMEGA;dw++){
				mpfr_set_d(A->data[dw],FINALJACKA[dw],GMP_RNDN);
			}

			//*********************************************************
			//	ADD THE NECESSARY DW FACTORS FOR THE INTEGRAL
			//*********************************************************

			append_dw(Adw,A,omega,NOMEGA,HPNUM,OMEGAHPMIN,OMEGAHPMAX);
		
			gmp_multmv(DCKernel,0,Adw,DA);
			gmp_multmv(Kernel,0,Adw,DAWCOR);

			//*********************************************************
			//	OUTPUT THE FINAL JACKNIFED SPECTRAL FUNCTION
			//	AND ITS DATA
			//*********************************************************
			
				sprintf(fname, "%s%s", RESULTNAME, "JACK_spec_rec.dat");
				rho_output=fopen(fname,"a");

				for(int w=0; w<NOMEGA; w++){
				fprintf(rho_output,"%d %d %e %e %e %e %e %e %e\n", DX,w,mpfr_get_d(omega->data[w],GMP_RNDN),GNUPR(FINALJACKA[w])/mpfr_get_d(mprior->data[w],GMP_RNDN),GNUPR(sqrt(FINALJACKVAR[w]))/mpfr_get_d(mprior->data[w],GMP_RNDN),GNUPR(FINALJACKA[w]),GNUPR(sqrt(FINALJACKVAR[w])),GNUPR(FINALJACKAERR[w])/mpfr_get_d(mprior->data[w],GMP_RNDN),GNUPR(FINALJACKAERR[w]));
				}
				fclose(rho_output);


				sprintf(fname, "%s%s", RESULTNAME, "JACK_spec_recdat.dat");
				rho_output=fopen(fname,"a");

				for(int t=0; t<TWIDTH; t++){
				fprintf(rho_output,"%d %d %e %e %e\n", DX,t,mpfr_get_d(tau->data[t],GMP_RNDN),GNUPR(mpfr_get_d(DA->data[t],GMP_RNDN)),GNUPR(mpfr_get_d(DAWCOR->data[t],GMP_RNDN)));
				}
				fclose(rho_output);
			
			
			gslgmp_vector_free(A);
			gslgmp_vector_free(Adw);
			gslgmp_vector_free(DA);
			gslgmp_vector_free(DAWCOR);

			free(FINALJACKA);
			free(FINALJACKVAR);
			for(int i=0;i<JACKNUM+1;i++) free(FINALAJACK[i]);
			for(int i=0;i<JACKNUM+1;i++) free(FINALERRJACK[i]);
			free(FINALAJACK);
			free(FINALERRJACK);
			free(FINALJACKAERR);

		}
		
		fprintf(stderr, "\n**********************************************************\n");
		fprintf(stderr, "***   SLICE DX:%d RECONSTRUCTION COMPLETE ALL DONE   ***\n",DX);
		fprintf(stderr, "**********************************************************\n");		
	}
		
	//*********************************************************************************
	//	END OF THE DX LOOP
	//*********************************************************************************

	fprintf(stderr, "\n**********************************************************\n");
	fprintf(stderr, "***   END OF PROGRAM!                                  ***\n");
	fprintf(stderr, "**********************************************************\n");
	
	return 1;

    }
